'''
Created on Feb 17, 2017

@author: calin
'''

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

class UnsupportedNodeTypeError(Exception):
    """ Unsupported node type failure. """
    def __init__(self, *args, **kwargs): # real signature unknown
        pass

    @staticmethod # known case of __new__
    def __new__(S, *more): # real signature unknown; restored from __doc__
        """ T.__new__(S, ...) -> a new object with type S, a subtype of T """
        pass


#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function


def is_classifier(cls):
    """
    Auxillary function to validate whether a given object is a
    classification-based model. (Can be used for the Fusion Handler as well as
    the Local Training Handler).

    :param cls: Object of interest to identify the type of model.
    :type  cls: `Object`
    :return `bool`
    """
    return cls.loss == 'auto' or cls.loss == 'binary_crossentropy' or \
        cls.loss == 'categorical_crossentropy'



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import time
import pandas as pd
import logging
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure
#from ibm_watson_machine_learning.tests.ICP.preparation_and_cleaning import *
from ibm_watson_machine_learning.wml_client_error import WMLClientError
from ibm_watson_machine_learning.tests.CP4D_35.preparation_and_cleaning import *
from ibm_watson_machine_learning.tests.CP4D_35.models_preparation import *


class TestWMLClientWithPMML(unittest.TestCase):
    deployment_id = None
    model_id = None
    scoring_url = None
    scoring_id = None
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithPMML.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()





        self.space_name = str(uuid.uuid4())

        metadata = {
            self.client.spaces.ConfigurationMetaNames.NAME: 'client_space_' + self.space_name,
            self.client.spaces.ConfigurationMetaNames.DESCRIPTION: self.space_name + ' description'
        }

        self.space = self.client.spaces.store(meta_props=metadata)

        TestWMLClientWithPMML.space_id = self.client.spaces.get_id(self.space)

        print("space_id: ", TestWMLClientWithPMML.space_id)
        self.client.set.default_space(TestWMLClientWithPMML.space_id)
        self.model_path = os.path.join(os.getcwd(), 'artifacts', 'pmml-model.xml')

        # TestWMLClientWithDO.logger.info("Service Instance: setting up credentials")
        # self.wml_credentials = get_wml_credentials()
        # self.client = get_client()
        # self.model_path = os.path.join(os.getcwd(), 'artifacts', 'DrugSelectionAutoAI_model_content.gzip')
        # self.update_model_path = os.path.join(os.getcwd(), 'artifacts', 'pipeline-model.json')

    # def test_01_set_space(self):
    #     space = self.client.spaces.store({self.client.spaces.ConfigurationMetaNames.NAME: "DO_test_case"})
    #
    #     TestWMLClientWithDO.space_id = self.client.spaces.get_id(space)
    #     self.client.set.default_space(TestWMLClientWithDO.space_id)
    #     self.assertTrue("SUCCESS" in self.client.set.default_space(TestWMLClientWithDO.space_id))


    def test_02_publish_hybrid_model_in_repository(self):
        TestWMLClientWithPMML.logger.info("Saving trained model in repo ...")
        TestWMLClientWithPMML.logger.debug("Model path: {}".format(self.model_path))

        self.client.repository.ModelMetaNames.show()

        sw_spec_id = self.client.software_specifications.get_id_by_name("spark-mllib_2.4")
        print(sw_spec_id)
        meta_props = {
            self.client.repository.ModelMetaNames.NAME: "pmmlmodel",
            self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: sw_spec_id,
            self.client.repository.ModelMetaNames.TYPE: 'pmml_4.2.1'}
        published_model = self.client.repository.store_model(self.model_path, meta_props=meta_props)
        TestWMLClientWithPMML.model_id = self.client.repository.get_model_id(published_model)
        TestWMLClientWithPMML.logger.info("Published model ID:" + str(TestWMLClientWithPMML.model_id))
        self.assertIsNotNone(TestWMLClientWithPMML.model_id)

    def test_04_get_details(self):
        TestWMLClientWithPMML.logger.info("Get details")
        self.assertIsNotNone(self.client.repository.get_details())
        det = self.client.repository.get_details()
        print(det)

    def test_05_get_model_details(self):
        TestWMLClientWithPMML.logger.info("Get model details")
        details = self.client.repository.get_details(TestWMLClientWithPMML.model_id)
        import json
        print(details)
        self.assertIsNotNone(self.client.repository.get_model_details(TestWMLClientWithPMML.model_id))

    def test_10_download_model(self):
        TestWMLClientWithPMML.logger.info("Download model")
        try:
            os.remove('download_test_url-pmml')
        except OSError:
            pass

        try:
            os.remove('download_test_url-pmml')
        except IOError:
            pass

        self.client.repository.download(TestWMLClientWithPMML.model_id, filename='download_test_url-pmml')
        try:
            os.remove('download_test_url-pmml')
        except OSError:
            pass

    def test_11_delete_model(self):
        TestWMLClientWithPMML.logger.info("Delete function")
        self.client.script.delete(TestWMLClientWithPMML.model_id)

    def test_12_delete_space(self):
        self.client.spaces.delete(TestWMLClientWithPMML.space_id)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .abstract_autoai_test import AbstractTestAutoAIAsync, AbstractTestAutoAISync
from .abstract_autoai_test_notebook_flow import AbstractTestAutoAISyncNotebookFlow
from .abstract_deployment_webservice import AbstractTestWebservice
from .abstract_deployment_batch import AbstractTestBatch
from .abstract_test_iris_wml_autoai_multiclass_connections import AbstractTestAutoAIRemote
from .abstract_test_iris_using_database_connection import AbstractTestAutoAIDatabaseConnection
from .abstract_test_iris_using_database_data_asset import AbstractTestAutoAIConnectedAsset
from .abstract_obm_autoai_test import AbstractTestOBM
from .abstract_timeseries_test import AbstractTestTSAsync



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import logging
# from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure
from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.tests.Cloud.preparation_and_cleaning import *
import os
import logging

logging.basicConfig(level=logging.DEBUG)


class TestFactsheets(unittest.TestCase):
    model_id = None
    model_content_path = f'{os.getcwd()}/Cloud/artifacts/tf_model_fvt_test.tar.gz'
    space_id = "c93565d9-41c2-4dbc-8586-2b32da8ad018"
    catalog_id = '12a02c34-0e5f-443a-9ea5-694ed6c6e56f'
    model_entry_id = None
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(cls):
        TestFactsheets.logger.info("Service Instance: setting up credentials")

        cls.wml_credentials = get_wml_credentials()
        cls.client = get_client()
        cls.client.set.default_space(TestFactsheets.space_id)

    def test_01_publish_local_model_in_repository(self):
        self.client.repository.ModelMetaNames.show()
        swspec_id = self.client.software_specifications.get_uid_by_name('default_py3.7')

        model_details = self.client.repository.store_model(
            model=self.model_content_path,
            meta_props={self.client.repository.ModelMetaNames.NAME: "test_tf",
                        self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: swspec_id,
                        self.client.repository.ModelMetaNames.TYPE: 'tensorflow_2.1'})
        TestFactsheets.model_id = self.client.repository.get_model_uid(model_details)

        print(TestFactsheets.model_id)

        self.assertIsNotNone(TestFactsheets.model_id)

    def test_02a__register_WKC_model_entry_new(self):
        register_details = self.client.factsheets.register_model_entry(
            model_id=self.model_id,
            meta_props={
                self.client.factsheets.ConfigurationMetaNames.NAME: 'Tensorflow model 2',
                self.client.factsheets.ConfigurationMetaNames.DESCRIPTION: 'Tensorflow model',
                self.client.factsheets.ConfigurationMetaNames.MODEL_ENTRY_CATALOG_ID: self.catalog_id
            })
        self.assertIsInstance(register_details, dict,
                              msg="register_details is not of dictionary type! It could be empty.")
        print(register_details)

        TestFactsheets.model_entry_id = register_details['model_entry_id']

    def test_02b__register_WKC_model_entry_existing(self):
        register_details = self.client.factsheets.register_model_entry(
            model_id=self.model_id,
            meta_props={
                self.client.factsheets.ConfigurationMetaNames.ASSET_ID: self.model_entry_id,
                self.client.factsheets.ConfigurationMetaNames.MODEL_ENTRY_CATALOG_ID: self.catalog_id
            })
        self.assertIsInstance(register_details, dict,
                              msg="register_details is not of dictionary type! It could be empty.")
        print(register_details)

    def test_03a_list_all_wkc_model_entries_for_specified_catalog(self):
        response = self.client.factsheets.list_model_entries(catalog_id=self.catalog_id)
        print(response)
        self.assertIsInstance(response, dict, msg="Response is not of dictionary type! It could be empty.")

    def test_03b_list_all_wkc_model_entries_for_all_catalogs(self):
        response = self.client.factsheets.list_model_entries()
        print(response)
        self.assertIsInstance(response, dict, msg="Response is not of dictionary type! It could be empty.")

    def test_04_unregister_WKC_model_entry(self):
        self.client.factsheets.unregister_model_entry(asset_id=self.model_id)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .library_imports import LibraryChecker
from ibm_watson_machine_learning.libs.repo.base_constants import *

lib_checker = LibraryChecker()

if lib_checker.installed_libs[PYSPARK]:
    from pyspark.ml import Pipeline, PipelineModel


class SparkUtil(object):
    DEFAULT_LABEL_COL = 'label'

    @staticmethod
    def get_label_col(spark_artifact):
        lib_checker.check_lib(PYSPARK)
        if isinstance(spark_artifact, PipelineModel):
            pipeline = Pipeline(stages=spark_artifact.stages)
            return SparkUtil.get_label_col_from__stages(pipeline.getStages())
        elif isinstance(spark_artifact, Pipeline):
            return SparkUtil.get_label_col_from__stages(spark_artifact.getStages())
        else:
            return SparkUtil.DEFAULT_LABEL_COL

    @staticmethod
    def get_label_col_from__stages(stages):
        lib_checker.check_lib(PYSPARK)
        label = SparkUtil._get_label_col_from_python_stages(stages)

        if label == SparkUtil.DEFAULT_LABEL_COL:
            label = SparkUtil._get_label_col_from_java_stages(stages)

        return label

    @staticmethod
    def _get_label_col_from_python_stages(stages):
        try:
            label_col = stages[-1].getLabelCol()
        except Exception as ex:
            label_col = SparkUtil.DEFAULT_LABEL_COL

        reversed_stages = stages[:]
        reversed_stages.reverse()

        for stage in reversed_stages[1:]:
            try:
                if stage.getOutputCol() == label_col:
                    label_col = stage.getInputCol()
            except Exception as ex:
                pass

        return label_col

    @staticmethod
    def _get_label_col_from_java_stages(stages):
        try:
            label_col = stages[-1]._call_java("getLabelCol")
        except Exception as ex:
            label_col = SparkUtil.DEFAULT_LABEL_COL

        reversed_stages = stages[:]
        reversed_stages.reverse()

        for stage in reversed_stages[1:]:
            try:
                if stage._call_java("getOutputCol") == label_col:
                    label_col = stage._call_java("getInputCol")
            except Exception as ex:
                pass

        return label_col



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import requests
import logging
from ibm_watson_machine_learning.wml_client_error import MissingValue, WMLClientError, NoWMLCredentialsProvided, ApiRequestFailure, UnexpectedType, MissingMetaProp
from ibm_watson_machine_learning.href_definitions import HrefDefinitions
from ibm_watson_machine_learning.utils import get_type_of_details, STR_TYPE, str_type_conv, NextResourceGenerator
import urllib.parse
import os
import sys

class WMLResource:
    def __init__(self, name, client):
        self._logger = logging.getLogger(__name__)
        self._name = name
        self._WSD = None
        WMLResource._validate_type(client, u'client', object, True)
        if client.wml_credentials is None:
            raise NoWMLCredentialsProvided
        WMLResource._validate_type(client.wml_credentials, u'wml_credentials', dict, True)
        if not client.WSD and not client.ICP:
            wml_token = str_type_conv(client.wml_token)
            WMLResource._validate_type(wml_token, u'wml_token', STR_TYPE, True)
        self._client = client
        self._ICP = client.ICP
        self._WSD = client.WSD

        # ToDO why this is differnt , check in 2.5 cluster and cloud whether it is working

        if (client.ICP or client.WSD) and not client.ICP_PLATFORM_SPACES:
            self._wml_credentials = client.wml_credentials
        else:
            self._wml_credentials = client.service_instance._wml_credentials

    def _handle_response(self, expected_status_code, operationName, response, json_response=True):
        if "dele" in operationName or "cancel migration" == operationName:
            if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                if response.status_code == expected_status_code:
                    return "SUCCESS"
                else:
                    msg = "{} failed. Reason: {} ".format(operationName, response.text)
                    raise WMLClientError(msg)
            else:
                if response.status_code == expected_status_code:
                    return "SUCCESS"
                else:
                    print(response.text)
                    return "FAILED"
        if response.status_code == expected_status_code:

            self._logger.info(u'Successfully finished {} for url: \'{}\''.format(operationName, response.url))
            self._logger.debug(u'Response({} {}): {}'.format(response.request.method, response.url, response.text))
            if json_response:
                try:
                    return response.json()
                except Exception as e:
                    raise WMLClientError(u'Failure during parsing json response: \'{}\''.format(response.text), e)
            else:
                return response.text
        else:
            raise ApiRequestFailure(u'Failure during {}.'.format(operationName), response)

    @staticmethod
    def _get_required_element_from_dict(el, root_path, path):
        WMLResource._validate_type(el, root_path, dict)
        WMLResource._validate_type(root_path, u'root_path', STR_TYPE)
        WMLResource._validate_type(path, u'path', list)

        if path.__len__() < 1:
            raise WMLClientError(u'Unexpected path length: {}'.format(path.__len__))

        try:
            new_el = el[path[0]]
            new_path = path[1:]
        except Exception as e:
            raise MissingValue(root_path + '.' + str(path[0]), e)

        if path.__len__() > 1:
            return WMLResource._get_required_element_from_dict(new_el, root_path + u'.' + path[0], new_path)
        else:
            if new_el is None:
                raise MissingValue(root_path + u'.' + str(path[0]))

            return new_el

    @staticmethod
    def _validate_type(el, el_name, expected_type, mandatory=True):
        if el_name is None:
            raise MissingValue(u'el_name')

        el_name = str_type_conv(el_name)
        if type(el_name) is not STR_TYPE:
            raise UnexpectedType(u'el_name', STR_TYPE, type(el_name))

        if expected_type is None:
            raise MissingValue(u'expected_type')

        if type(expected_type) is not type and type(expected_type) is not list:
            raise UnexpectedType('expected_type', 'type or list', type(expected_type))

        if type(mandatory) is not bool:
            raise UnexpectedType(u'mandatory', bool, type(mandatory))

        if mandatory and el is None:
            raise MissingValue(el_name)
        elif el is None:
            return

        if type(expected_type) is list:
            try:
                next((x for x in expected_type if isinstance(el, x)))
                return True
            except StopIteration:
                return False
        else:
            if not isinstance(el, expected_type):
                raise UnexpectedType(el_name, expected_type, type(el))

    @staticmethod
    def _validate_meta_prop(meta_props, name, expected_type, mandatory=True):
        if name in meta_props:
            WMLResource._validate_type(meta_props[name], u'meta_props.' + name, expected_type, mandatory)
        else:
            if mandatory:
                raise MissingMetaProp(name)

    @staticmethod
    def _validate_type_of_details(details, expected_type):
        actual_type = get_type_of_details(details)

        if type(expected_type) is list:
            expected_types = expected_type
        else:
            expected_types = [expected_type]

        if not any([actual_type == exp_type for exp_type in expected_types]):
            logger = logging.getLogger(__name__)
            logger.debug(u'Unexpected type of \'{}\', expected: \'{}\', actual: \'{}\', occured for details: {}'.format(
                u'details', expected_type, actual_type, details))
            raise UnexpectedType(u'details', expected_type, actual_type)

    def _check_if_lib_or_def(self, lib_uid):

        verify = False
        lib_details = requests.get(self._client.service_instance._href_definitions.get_custom_library_href(lib_uid),
                                   headers=self._client._get_headers(), verify=verify)
        if lib_details.status_code == 200:
            return lib_details.json()["metadata"]["href"]
        def_details = requests.get(self._client.service_instance._href_definitions.get_model_definition_assets_href() + u'/' + lib_uid,
                                   params=self._client._params(), headers=self._client._get_headers(), verify=verify)
        if def_details.status_code == 200:
            return def_details.json()["href"]
        else:
            raise WMLClientError("The LIB_UID provided is not a valid model_definition/library")

    def _get_artifact_details(self, base_url, uid, limit, resource_name, summary=None,
                              pre_defined=None, query_params=None, _async=False, _all=False):
        op_name = 'getting {} details'.format(resource_name)

        if uid is None:
            return self._get_with_or_without_limit(
                url=base_url,
                limit=limit,
                op_name=resource_name,
                summary=summary,
                pre_defined=pre_defined,
                query_params=query_params,
                _async=_async,
                _all=_all
            )
        else:
            # if not is_uid(uid):
            #     raise WMLClientError(u'Failure during {}, invalid uid: \'{}\''.format(op_name, uid)) # TODO
            if query_params is None:
                params = self._client._params()
            else:
                params = query_params

            if 'userfs' in params:
                params.pop('userfs')

            url = base_url + u'/' + uid

            if not self._ICP and not self._client.WSD:
                response_get = requests.get(
                    url,
                    params,
                    headers=self._client._get_headers()
                )
            else:
                response_get = requests.get(
                    url,
                    params,
                    headers=self._client._get_headers(),
                    verify=False
                )

            return self._handle_response(200, op_name, response_get)

    def _get_with_or_without_limit(self, url, limit, op_name, summary, pre_defined, revision=None,
                                   skip_space_project_chk=False, query_params=None, _async=False, _all=False):
        params = self._client._params(skip_space_project_chk)
        if query_params is not None:
            params.update(query_params)

        if summary == "False":
            params.update({u'summary': u'false'})

        if pre_defined == "True":
            params.update({u'system_runtimes': u'true'})

        if limit is not None:
            if limit < 1:
                raise WMLClientError('Limit cannot be lower than 1.')
            elif limit > 200:
                raise WMLClientError('Limit cannot be larger than 200.')
            params.update({u'limit': limit})
        else:
            params.update({u'limit': 200})

        if revision is not None:
            if op_name == 'asset_revision':
                params.update({u'revision_id': revision}) # CAMS assets api takes 'revision_id' query parameter
            else:
                params.update({u'rev': revision})

        resources = []

        href = "/".join(url.split('/')[3:])
        url_2 = "/".join(url.split('/')[:3])
        resource_generator = NextResourceGenerator(self._client, url_2, href, params, _all)

        if _async:
            return ({"resources": entry} for entry in resource_generator)

        else:
            if _all:
                for entry in resource_generator:
                    resources.extend(entry)
                return {
                    "resources": resources
                }

            else:
                if not self._ICP:
                    response_get = requests.get(
                        url,
                        headers=self._client._get_headers(),
                        params=params
                    )
                else:
                    response_get = requests.get(
                        url,
                        headers=self._client._get_headers(),
                        params=params,
                        verify=False
                    )

                result = self._handle_response(200, op_name, response_get)
                if 'resources' in result:
                    resources.extend(result['resources'])

                elif 'metadata' in result:
                    resources.extend([result])

                else:
                    resources.extend(result.get('results'))

                return {
                    "resources": resources
                }

    def _get_no_space_artifact_details(self, base_url, uid, limit, resource_name, summary=None,pre_defined=None):
        op_name = 'getting {} details'.format(resource_name)

        if uid is None:
            return self._get_no_space_with_or_without_limit(url=base_url, limit=limit, op_name=resource_name,summary=summary,pre_defined=pre_defined)
        else:
            # if not is_uid(uid):
            #     raise WMLClientError(u'Failure during {}, invalid uid: \'{}\''.format(op_name, uid)) # TODO

            url = base_url + u'/' + uid

            if not self._ICP:
                response_get = requests.get(
                    url,
                    headers=self._client._get_headers()
                )
            else:
                response_get = requests.get(
                    url,
                    headers=self._client._get_headers(),
                    verify=False
                )

            return self._handle_response(200, op_name, response_get)

    def _get_no_space_with_or_without_limit(self, url, limit, op_name,summary,pre_defined):
        params = {}
        if summary == "False":
            params.update({u'summary': u'false'})
        if pre_defined == "True":
            params.update({u'system_runtimes': u'true'})
        if limit is not None:
            if limit < 1:
                raise WMLClientError('Limit cannot be lower than 1.')
            elif limit > 1000:
                raise WMLClientError('Limit cannot be larger than 1000.')
            params.update({u'limit': limit})
        if len (params) > 0:
            if not self._ICP:
                response_get = requests.get(
                    url,
                    headers=self._client._get_headers(),
                    params=params
                )
            else:
                response_get = requests.get(
                    url,
                    headers=self._client._get_headers(),
                    params=params,
                    verify=False
                )

            return self._handle_response(200, op_name, response_get)
        else:

            resources = []

            while True:
                if not self._ICP:
                    response_get = requests.get(
                        url,
                        headers=self._client._get_headers(),

                    )
                else:
                    response_get = requests.get(
                        url,
                        headers=self._client._get_headers(),
                        verify=False
                    )

                result = self._handle_response(200, op_name, response_get)
                resources.extend(result['resources'])

                if 'next' not in result:
                    break
                else:

                    url = self._wml_credentials["url"]+result['next']['href']
                    if('start=invalid' in url):
                        break
            return {
                "resources": resources
            }

    def _if_deployment_exist_for_asset(self, asset_id):

        params = self._client._params()
        if 'project_id' in params.keys():
            return 0
        deployment_href = self._client.service_instance._href_definitions.get_deployments_href() + "?asset_id=" + asset_id
        response_deployment = requests.get(deployment_href, params=self._client._params(),
                                           headers=self._client._get_headers(), verify=False)
        deployment_json = self._handle_response(200, "Get deployment details", response_deployment)
        resources = deployment_json["resources"]
        if resources:
            return 1
        else:
            return 0

    def _list(self, values, header, limit, default_limit, sort_by='CREATED'):
        if sort_by is not None and sort_by in header:
            column_no = header.index(sort_by)
            values = sorted(values, key=lambda x: x[column_no], reverse=True)

        from tabulate import tabulate

        if limit is None:
            table = tabulate([header] + values[:default_limit])
            print(table)
            if default_limit is not None and len(values) > default_limit:
                print('Note: Only first {} records were displayed. To display more use \'limit\' parameter.'.format(default_limit))
        else:
            table = tabulate([header] + values[:limit])
            print(table)

        try:
            import pandas as pd
            if limit is None:
                return pd.DataFrame(values[:default_limit], columns=header)

            else:
                return pd.DataFrame(values[:limit], columns=header)

        except ImportError:
            if limit is None:
                return [header] + values[:default_limit]

            else:
                return [header] + values[:limit]

    def _wsd_list_assets(self, asset_type, limit=None):
        _DEFAULT_LIST_LENGTH = 50
        ##For CP4D, check if either spce or project ID is set
        href = self._client.service_instance._href_definitions.get_wsd_asset_search_href(asset_type)
        if limit is None:
            data = {
                "query": "*:*"
            }
        else:
            self._validate_type(limit, u'limit', int, False)
            data = {
                "query": "*:*",
                "limit": limit
            }

        response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(),
                                     json=data, verify=False)
        self._handle_response(200, u'list ', response)
        asset_details = self._handle_response(200, u'model list', response)["results"]
        model_def_values = [
            (m[u'metadata'][u'name'], m[u'metadata'][u'asset_type'], m[u'metadata'][u'asset_id']) for
            m in asset_details]

        self._list(model_def_values, [u'NAME', u'ASSET_TYPE', u'GUID'], limit, _DEFAULT_LIST_LENGTH)

    def _create_wsd_payload(self, asset_type, payload_input, meta_props):
        # TODO: Get the country code from locale
        origin_country = "US"
        payload_metadata = {
            "name": meta_props['name'],
            "asset_type": asset_type,
            "origin_country": origin_country,
            "assetCategory": "USERS"
        }
        if 'description' in meta_props and meta_props['description'] is not None:
            payload_metadata.update({'description': meta_props['description']})

        if asset_type == 'wml_model':
            payload_input.update({'content_status': {
                "state": "persisted"}})
            if 'trainingDataSchema' in meta_props and meta_props['trainingDataSchema'] is not None:
                training_schema_field = meta_props['trainingDataSchema']['fields']
                training_data_reference=[
                    {'location': {'bucket': 'not_applicable'},
                     'type': 'fs',
                     'connection': {'access_key_id': 'not_applicable',
                                    'secret_access_key': 'not_applicable',
                                    'endpoint_url': 'not_applicable'},
                     'schema': {
                         'id': '1',
                         'type': 'struct',
                         'fields': training_schema_field
                     }}]
                payload_input.pop('trainingDataSchema')
                payload_input.update({'training_data_references': training_data_reference})
        cams_payload = {
            "metadata": payload_metadata,
            "entity": {
                asset_type: payload_input
            }
        }
        return cams_payload

    def _wsd_create_asset(self, asset_type, input_payload, meta_props, path_to_archive, user_archive_file=False):
        import json
        import copy
        def is_xml(model_filepath):
            return os.path.splitext(os.path.basename(model_filepath))[-1] == '.xml'

        params = self._client._params()
        headers = self._client._get_headers()
        url = self._client.service_instance._href_definitions.get_wsd_model_href()

        if asset_type != 'wml_model_definition':
            cams_entity = copy.deepcopy(input_payload)
            #cams_entity.pop('name')
            if cams_entity.get('description') is not None:
                cams_entity.pop('description')
            cams_payload = self._create_wsd_payload(asset_type, cams_entity, meta_props)
        else:
            cams_payload = copy.deepcopy(input_payload)
        try:
            atype_body = {
                "name": asset_type
            }
            atype_payload = json.dumps(atype_body, separators=(',', ':'))
            atype_url = self._client.service_instance._href_definitions.get_wsd_asset_type_href()
            aheaders = {
                'Content-Type': "application/json"
            }
            asset_type_response = requests.post(
                atype_url,
                params=params,
                data=atype_payload,
                headers=aheaders,
                verify=False
            )

            if asset_type_response.status_code != 200 and \
                    asset_type_response.status_code != 201 and asset_type_response.status_code != 409:
                raise WMLClientError("Failed to create asset type. Try again.")

            create_response = requests.post(
                url,
                params=params,
                json=cams_payload,
                headers=headers,
                verify=False
            )
            asset_res_text = 'creating new ' + asset_type[4:]
            asset_details = self._handle_response(201, asset_res_text, create_response)
            # Upload model content to desktop project using polyfill
            if create_response.status_code == 201:
                asset_uid = create_response.json()["metadata"]["asset_id"]

                if type(path_to_archive) is STR_TYPE:
                    file_name_to_attach = path_to_archive.rsplit("/")[-1]
                else:
                    file_name_to_attach = 'function_attachment'
                if asset_type == 'data_asset':
                    content_upload_url = self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + \
                                         urllib.parse.quote(asset_type + "/" + input_payload['name'],safe='')
                else:
                    content_upload_url = self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + \
                                     urllib.parse.quote(asset_type + "/" + asset_uid + "/" + file_name_to_attach, safe='')
                attach_url = self._client.service_instance._href_definitions.get_model_definition_assets_href() + "/" + \
                             urllib.parse.quote(asset_uid + "/attachments")
                if type(path_to_archive) is STR_TYPE:
                    with open(path_to_archive, 'rb') as f:
                        fdata = f.read()
                else:
                    fdata = path_to_archive

                if is_xml(path_to_archive):
                    response = requests.put(
                        content_upload_url,
                        files={'file': ('native', fdata, 'application/xml', {'Expires': '0'})},
                        params=params,
                        verify=False
                    )
                else:
                    response = requests.put(
                        content_upload_url,
                        files={'file': ('native', fdata, 'application/gzip', {'Expires': '0'})},
                        params=params,
                        verify=False
                    )

                if response.status_code == 201:
                    # update the attachement url with details :
                    if asset_type == 'data_asset':
                        asset_body = {
                            "asset_type": asset_type,
                            "name": input_payload['name'],
                            "mime": input_payload['mime_type'],
                            "object_key": input_payload['name'],
                            "object_key_is_read_only": True
                        }
                    else:
                        asset_body = {
                            "asset_type": asset_type,
                            "name": "native",
                            "object_key": asset_uid + "/" + file_name_to_attach,
                            "object_key_is_read_only": True
                        }
                    attach_payload = json.dumps(asset_body, separators=(',', ':'))

                    attach_response = requests.post(attach_url,
                                                    data=attach_payload,
                                                    params=params,
                                                    headers=headers,
                                                    verify=False)
                    if attach_response.status_code == 201:
                        # remove the tmp file create for content upload
                        if asset_type != 'data_asset':
                            if type(path_to_archive) is STR_TYPE and os.path.isfile(path_to_archive) \
                                    and user_archive_file is False:
                                os.remove(path_to_archive)
                            return self._wsd_get_required_element_from_response(asset_details)
                        else:
                            return asset_details
                    else:
                        self._wsd_delete_asset(asset_uid)
                        raise WMLClientError("Failed while creating a model. Try again.")
                else:
                    self._wsd_delete_asset(asset_uid)
                    raise WMLClientError("Failed while creating a model. Try again.")
            else:
                raise WMLClientError("Failed while creating a model. Try again.")
        except Exception as e:
            raise e

    def _wsd_delete_asset(self, asset_type, asset_id):
        asset_uid = str_type_conv(asset_id)
        delete_endpoint = self._client.service_instance._href_definitions.get_model_definition_assets_href() + "/" + asset_uid
        self._logger.debug(u'Deletion artifact {} endpoint: {}'.format(asset_type, delete_endpoint))
        response_delete = requests.delete(delete_endpoint, params=self._client._params(),
                                              headers=self._client._get_headers(), verify=False)
        return response_delete

    def _wsd_get_asset_details(self, asset_type, asset_id, limit=None):
        url = self._client.service_instance._href_definitions.get_model_definition_assets_href()
        return self._get_artifact_details(url, asset_id, limit, asset_type)

    def _wsd_get_required_element_from_response(self, response_data):
        WMLResource._validate_type(response_data, u'model response', dict)
        try:
            if self._client.default_project_id is not None and self._client.WSD:
                href = self._client.service_instance._href_definitions.get_base_asset_with_type_href(
                    response_data['metadata']['asset_type'],
                    response_data['metadata']['asset_id']) + "?" + "project_id=" + response_data['metadata']['project_id']
                metadata_val = {'project_id': response_data['metadata']['project_id'],
                                       'asset_id': response_data['metadata']['asset_id'],
                                       'href': href,
                                       'name': response_data['metadata']['name'],
                                       'asset_type': response_data['metadata']['asset_type'],
                                       'created_at': response_data['metadata']['created_at'],
                                       'last_updated_at': response_data['metadata']['usage']['last_updated_at']
                                       }
                if 'description' in response_data['metadata'] and response_data['metadata']['description'] is not None:
                    metadata_val.update({u'description': response_data['metadata']['description']})

                new_el = {'metadata': metadata_val,
                          'entity': response_data['entity']
                          }
                if 'frameworkName' in new_el['entity']:
                     new_el['entity'].pop('frameworkName')
                     new_el['entity'].pop('frameworkVersion')
                if 'framework_runtimes' in new_el['entity']:
                    new_el['entity'].pop('framework_runtimes')
                return new_el
            else:
                return response_data
        except Exception as e:
            raise WMLClientError("Failed to read Response from down-stream service: " + response_data)

    def _create_revision_artifact(self, base_url, uid, resource_name):
        op_name = 'Creation revision for {}'.format(resource_name)
        if self._client.default_project_id is not None:
            input_json = {
                "project_id": self._client.default_project_id
            }
        else:
            input_json = {
                "space_id": self._client.default_space_id
            }

        url = base_url + u'/' + uid + "/revisions"
        if not self._ICP and not self._client.WSD:
            if self._client.CLOUD_PLATFORM_SPACES:
                params = self._client._params(skip_for_create=True)
                response = requests.post(url, headers=self._client._get_headers(),
                                         params=params,
                                         json=input_json)
            else:
                response = requests.post(url, headers=self._client._get_headers(),
                                         json=input_json)
        else:
            if self._client.ICP_PLATFORM_SPACES:
                response = requests.post(url, headers=self._client._get_headers(),
                                         params=self._client._params(skip_for_create=True),
                                         json=input_json, verify=False)
            else:
                response = requests.post(url, headers=self._client._get_headers(),
                                         json=input_json, verify=False)

        return self._handle_response(201, op_name, response)

    def _create_revision_artifact_for_assets(self, uid, resource_name):
        op_name = 'Creation revision for {}'.format(resource_name)

        url = self._client.service_instance._href_definitions.get_asset_href(uid) + "/revisions"
        commit_message = "Revision creation for " + resource_name + " " + uid

        payload_json = {
            "commit_message": commit_message
        }

        # CAMS revision creation api takes space_id as a query parameter. Hence
        # params has to be passed

        if not self._ICP and not self._client.WSD:
            response = requests.post(url,
                                     headers=self._client._get_headers(),
                                     params=self._client._params(),
                                     json=payload_json)
        else:
            response = requests.post(url,
                                     headers=self._client._get_headers(),
                                     params=self._client._params(),
                                     json=payload_json,
                                     verify=False)

        if response.status_code == 201:
            print("DONE")

        return self._handle_response(201, op_name, response)

    def _update_attachment_for_assets(self,
                                      asset_type,
                                      asset_id,
                                      file_path,
                                      current_attachment_id=None):

        if current_attachment_id is not None:
            # Delete existing attachment to upload new attachment
            attachments_id_url = self._client.service_instance._href_definitions.get_asset_href(
                asset_id) + '/attachments/' + current_attachment_id

            if not self._ICP:
                delete_attachment_response = requests.delete(
                    attachments_id_url,
                    headers=self._client._get_headers(),
                    params=self._client._params()
                )
            else:
                delete_attachment_response = requests.delete(
                    attachments_id_url,
                    headers=self._client._get_headers(),
                    params=self._client._params(),
                    verify=False
                )

        if delete_attachment_response.status_code == 204 or current_attachment_id is None:
            attachment_meta = {
                "asset_type": asset_type,
                "name": "attachment_" + asset_id
            }

            attachments_url = self._client.service_instance._href_definitions.get_asset_href(asset_id) + '/attachments'

            # STEP 3b.
            # Get the signed url from CAMS to upload the attachment
            if not self._ICP:
                attachment_response = requests.post(
                    attachments_url,
                    headers=self._client._get_headers(),
                    params=self._client._params(),
                    json=attachment_meta
                )
            else:
                attachment_response = requests.post(
                    attachments_url,
                    headers=self._client._get_headers(),
                    json=attachment_meta,
                    params=self._client._params(),
                    verify=False
                )

            attachment_details = self._handle_response(201, u'creating new attachment', attachment_response)
            if attachment_response.status_code == 201:

                file = {'file': open(file_path, 'rb')}

                attachment_id = attachment_details["attachment_id"]
                attachment_url = attachment_details["url1"]

                # STEP 3c.
                # Upload attachment
                if not self._ICP:
                    put_response = requests.put(
                        attachment_url,
                        data=open(file_path, 'rb').read()
                    )
                else:
                    put_response = requests.put(
                        self._wml_credentials['url'] + attachment_url,
                        files=file,
                        verify=False
                    )

                if put_response.status_code == 201 or put_response.status_code == 200:
                    # STEP 3d.
                    # Mark attachment complete
                    if not self._ICP:
                        complete_response = requests.post(
                            self._client.service_instance._href_definitions.get_attachment_href(asset_id, attachment_id) + "/complete",
                            headers=self._client._get_headers(),
                            params=self._client._params()

                        )
                    else:
                        complete_response = requests.post(
                            self._client.service_instance._href_definitions.get_attachment_href(asset_id, attachment_id) + "/complete",
                            headers=self._client._get_headers(),
                            params=self._client._params(),
                            verify=False
                        )

                    if complete_response.status_code == 200:
                        print("SUCCESS")
                    else:
                        self._logger.error('Error in marking attachment complete for asset {}'.format(asset_id))
                        return "error_in_marking_attachment_complete"
                else:
                    self._logger.error('Error in uploading attachment for asset {}'.format(asset_id))
                    return "error_in_uploading_attachment"
            else:
                self._logger.error('Error in getting signed url for attachment for asset {}'.format(asset_id))
                return "error_in_getting_signed_url"
        else:
            self._logger.error('Error in deleting existing attachment {} for asset {}'.format(current_attachment_id,
                                                                                              asset_id))
            return "error_in_deleting_existing_attachment"

        return "success"

    def _chk_and_block_create_update_for_python36(self):
        pass
        # op = traceback.extract_stack(None, 2)[0][2]

        # if (3 == sys.version_info.major) and (6 == sys.version_info.minor):
        #     if self._client.CLOUD_PLATFORM_SPACES or self._client.CLOUD_BETA_FLOW:
        #         raise WMLClientError("Action not allowed!! Python 3.6 framework is deprecated and will be removed on Jan 20th, 2021. "
        #               "It will be read-only mode starting Nov 20th, 2020. i.e, you won't be able to create or update new assets and deployment using this client. "
        #               "Use Python 3.7 instead. For details, see https://dataplatform.cloud.ibm.com/docs/content/wsj/analyze-data/pm_service_supported_frameworks.html")




# coding: utf-8

"""

    No descripton provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pprint import pformat
from six import iteritems
import re


class HyperParametersExperiments(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, name=None, double_values=None, int_values=None, string_values=None, double_range=None, int_range=None):
        """
        HyperParametersExperiments - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'name': 'str',
            'double_values': 'list[float]',
            'int_values': 'list[int]',
            'string_values': 'list[str]',
            'double_range': 'HyperParametersExperimentsDoubleRange',
            'int_range': 'HyperParametersExperimentsIntRange'
        }

        self.attribute_map = {
            'name': 'name',
            'double_values': 'double_values',
            'int_values': 'int_values',
            'string_values': 'string_values',
            'double_range': 'double_range',
            'int_range': 'int_range'
        }

        self._name = name
        self._double_values = double_values
        self._int_values = int_values
        self._string_values = string_values
        self._double_range = double_range
        self._int_range = int_range

    @property
    def name(self):
        """
        Gets the name of this HyperParametersExperiments.
        Name of the parameter which we want to tune

        :return: The name of this HyperParametersExperiments.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this HyperParametersExperiments.
        Name of the parameter which we want to tune

        :param name: The name of this HyperParametersExperiments.
        :type: str
        """

        self._name = name

    @property
    def double_values(self):
        """
        Gets the double_values of this HyperParametersExperiments.


        :return: The double_values of this HyperParametersExperiments.
        :rtype: list[float]
        """
        return self._double_values

    @double_values.setter
    def double_values(self, double_values):
        """
        Sets the double_values of this HyperParametersExperiments.


        :param double_values: The double_values of this HyperParametersExperiments.
        :type: list[float]
        """

        self._double_values = double_values

    @property
    def int_values(self):
        """
        Gets the int_values of this HyperParametersExperiments.


        :return: The int_values of this HyperParametersExperiments.
        :rtype: list[int]
        """
        return self._int_values

    @int_values.setter
    def int_values(self, int_values):
        """
        Sets the int_values of this HyperParametersExperiments.


        :param int_values: The int_values of this HyperParametersExperiments.
        :type: list[int]
        """

        self._int_values = int_values

    @property
    def string_values(self):
        """
        Gets the string_values of this HyperParametersExperiments.


        :return: The string_values of this HyperParametersExperiments.
        :rtype: list[str]
        """
        return self._string_values

    @string_values.setter
    def string_values(self, string_values):
        """
        Sets the string_values of this HyperParametersExperiments.


        :param string_values: The string_values of this HyperParametersExperiments.
        :type: list[str]
        """

        self._string_values = string_values

    @property
    def double_range(self):
        """
        Gets the double_range of this HyperParametersExperiments.


        :return: The double_range of this HyperParametersExperiments.
        :rtype: HyperParametersExperimentsDoubleRange
        """
        return self._double_range

    @double_range.setter
    def double_range(self, double_range):
        """
        Sets the double_range of this HyperParametersExperiments.


        :param double_range: The double_range of this HyperParametersExperiments.
        :type: HyperParametersExperimentsDoubleRange
        """

        self._double_range = double_range

    @property
    def int_range(self):
        """
        Gets the int_range of this HyperParametersExperiments.


        :return: The int_range of this HyperParametersExperiments.
        :rtype: HyperParametersExperimentsIntRange
        """
        return self._int_range

    @int_range.setter
    def int_range(self, int_range):
        """
        Sets the int_range of this HyperParametersExperiments.


        :param int_range: The int_range of this HyperParametersExperiments.
        :type: HyperParametersExperimentsIntRange
        """

        self._int_range = int_range

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import io
import abc
import time
import contextlib

from ibm_watson_machine_learning.tests.base.abstract.abstract_deployment_test import AbstractDeploymentTest



class AbstractBatchDeploymentTest(AbstractDeploymentTest, abc.ABC):
    """
    Abstract class implementing scoring using batch deployment jobs.
    """
    def create_deployment_props(self):
        return {
            self.wml_client.deployments.ConfigurationMetaNames.NAME: self.deployment_name,
            self.wml_client.deployments.ConfigurationMetaNames.BATCH: {},
            self.wml_client.deployments.ConfigurationMetaNames.HARDWARE_SPEC: {"name": "S", "num_nodes": 1}
        }

    def test_09_download_deployment(self):
        pass

    def test_10_score_deployments(self):
        job_payload = self.create_scoring_payload()
        job_details = self.wml_client.deployments.create_job(self.deployment_id, job_payload)

        AbstractBatchDeploymentTest.job_id = self.wml_client.deployments.get_job_uid(job_details)
        self.assertIsNotNone(self.job_id)

    def test_10b_list_jobs(self):
        with io.StringIO() as buf, contextlib.redirect_stdout(buf):
            self.wml_client.deployments.list_jobs()
            jobs_list = buf.getvalue()

        self.assertIn(self.job_id, jobs_list)

    def test_10b_get_job_details(self):
        status = None
        elapsed_time = 0
        wait_time = 5
        max_wait_time = 500
        while status not in ['completed', 'failed'] and elapsed_time < max_wait_time:
            time.sleep(wait_time)
            elapsed_time += wait_time
            status = self.wml_client.deployments.get_job_status(self.job_id).get('state')
        self.assertEqual(status, "completed")



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import requests
from ibm_watson_machine_learning.utils import SPACES_DETAILS_TYPE, INSTANCE_DETAILS_TYPE, MEMBER_DETAILS_TYPE,DATA_ASSETS_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, meta_props_str_conv, str_type_conv, get_file_from_cos
from ibm_watson_machine_learning.metanames import VolumeMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure
import os
import shlex
import subprocess
_DEFAULT_LIST_LENGTH = 50


class Volume(WMLResource):
    """
    Store and manage your scripts assets.

    """
    ConfigurationMetaNames = VolumeMetaNames()
    """MetaNames for script Assets creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, volume_id):
        """
            Get Volume  details.

            **Parameters**

            .. important::
                #. **volume_id**: Unique ID of the volume.\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: Metadata of the volume details \n
                **return type**: dict\n

            **Example**

             >>> volume_details = client.volumes.get_details(volume_id)

        """
        Volume._validate_type(volume_id, u'volume_id', STR_TYPE, True)

        params = {'addon_type': 'volumes',
                  'include_service_status':True
                  }
        if not self._ICP:
            response = requests.get(self._client.service_instance._href_definitions.volume_href(volume_id),
                                    headers=self._client._get_headers(zen=True))
        else:
            response = requests.get(self._client.service_instance._href_definitions.volume_href(volume_id),
                                      headers=self._client._get_headers(zen=True), verify=False)
        if response.status_code == 200:
            return response.json()
        else:
            print(response.status_code, response.text)
            raise WMLClientError("Failed to Get the volume details. Try again.")


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create(self, meta_props):
        """
                Creates a Volume asset.

                **Parameters**

                .. important::
                   #. **meta_props**:  Name to be given to the Volume asset\n

                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: metadata of the created volume details\n
                    **return type**: dict\n

                **Example**
                Provision new PVC volume :
                 >>> metadata = {
                 >>>        client.volumes.ConfigurationMetaNamess.NAME: 'volume-for-wml-test',
                 >>>        client.volumes.ConfigurationMetaNames.NAMESPACE: 'wmldev2',
                 >>>        client.volumes.ConfigurationMetaNames.STORAGE_CLASS: 'nfs-client'
                 >>>        client.volumes.ConfigurationMetaNames.STORAGE_SIZE: "2G"
                 >>>    }
                 >>>
                 >>> asset_details = client.scripts.store(meta_props=metadata)

               Provision a Existing PVC volume:

                 >>> metadata = {
                 >>>        client.volumes.ConfigurationMetaNamess.NAME: 'volume-for-wml-test',
                 >>>        client.volumes.ConfigurationMetaNames.NAMESPACE: 'wmldev2',
                 >>>        client.volumes.ConfigurationMetaNames.EXISTING_PVC_NAME: 'volume-for-wml-test'
                 >>>    }
                 >>>
                 >>> asset_details = client.scripts.store(meta_props=metadata)

        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Failed to create volume. It is supported only for CP4D 3.5")

        volume_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client
        )

        create_meta = {}
        if self.ConfigurationMetaNames.EXISTING_PVC_NAME in meta_props and \
            meta_props[self.ConfigurationMetaNames.EXISTING_PVC_NAME] is not None:
            if self.ConfigurationMetaNames.STORAGE_CLASS in meta_props and \
                    meta_props[self.ConfigurationMetaNames.STORAGE_CLASS] is not None:
                raise WMLClientError("Failed while creating volume. Either provide EXISTING_PVC_NAME to create a volume using existing volume or"
                                     "provide STORAGE_CLASS and STORAGE_SIZE for new volume creation")
            else:
                create_meta.update({ "existing_pvc_name": meta_props[self.ConfigurationMetaNames.EXISTING_PVC_NAME]})
        else:
            if self.ConfigurationMetaNames.STORAGE_CLASS in meta_props and \
               meta_props[self.ConfigurationMetaNames.STORAGE_CLASS] is not None:
               if self.ConfigurationMetaNames.STORAGE_SIZE in meta_props and \
                       meta_props[self.ConfigurationMetaNames.STORAGE_SIZE] is not None:
                   create_meta.update({"storageClass": meta_props[self.ConfigurationMetaNames.STORAGE_CLASS]})
                   create_meta.update({"storageSize": meta_props[self.ConfigurationMetaNames.STORAGE_SIZE]})
               else:
                   raise WMLClientError("Failed to create volume. Missing input STORAGE_SIZE" )

        if self.ConfigurationMetaNames.EXISTING_PVC_NAME in meta_props and meta_props[self.ConfigurationMetaNames.EXISTING_PVC_NAME] is not None:
            input_meta = {
                "addon_type":"volumes",
                "addon_version":"-",
                "create_arguments":{
                    "metadata":create_meta
                },
                "namespace":meta_props[self.ConfigurationMetaNames.NAMESPACE],
                "display_name":meta_props[self.ConfigurationMetaNames.NAME]
            }
        else:
            input_meta = {
                "addon_type": "volumes",
                "addon_version": "-",
                "create_arguments": {
                    "metadata": create_meta
                },
                "namespace": meta_props[self.ConfigurationMetaNames.NAMESPACE],
                "display_name": meta_props[self.ConfigurationMetaNames.NAME]
            }
        creation_response = {}
        try:
            if self._client.CLOUD_PLATFORM_SPACES:
                creation_response = requests.post(
                    self._client.service_instance._href_definitions.volumes_href(),
                    headers=self._client._get_headers(zen=True),
                    json=input_meta
                )

            else:
                creation_response = requests.post(self._client.service_instance._href_definitions.volumes_href(),
                        headers=self._client._get_headers(zen=True),
                        json=input_meta,
                        verify=False
                    )
            if creation_response.status_code == 200:
                volume_id_details = creation_response.json()
                import copy
                volume_details = copy.deepcopy(input_meta)
                volume_details.update(volume_id_details)
                return volume_details
            else:
                print(creation_response.status_code, creation_response.text)
                raise WMLClientError("Failed to create a volume. Try again.")
        except Exception as e:
            print("Exception: ", {e})
            raise WMLClientError("Failed to create a volume. Try again.")



    def start(self, name):
        """
            Start the  volume service.

            **Parameters**

            .. important::
                #. **volume_name**:  Unique name of the volume to be started
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

                >>> client.volumes.start(volume_name)

        """

        WMLResource._chk_and_block_create_update_for_python36(self)
        if not self._client.ICP_PLATFORM_SPACES and not self._client.CLOUD_PLATFORM_SPACES:
            raise WMLClientError("Volume APIs are not supported. It is supported only for CP4D 3.5")
        start_url = self._client.service_instance._href_definitions.volume_service_href(name)
        # Start the volume  service
        start_data = {}
        try:
            if not self._ICP:
                start_data = {}
                creation_response = requests.post(
                    start_url,
                    headers=self._client._get_headers(zen=True),
                    json=start_data
                )
            else:
                creation_response = requests.post(
                   start_url,
                    headers=self._client._get_headers(zen=True),
                    json=start_data,
                    verify=False
                )
            if creation_response.status_code == 200:
                print("Volume Service started")
            elif creation_response.status_code == 500:
                print("Failed to start the volume. Make sure volume is in running with status RUNNING or UNKNOW and then re-try")
            else:
                print(creation_response.status_code, creation_response.text)
                raise WMLClientError("Failed to start the file to  volume. Try again.")
        except Exception as e:
            print("Exception:", {e})
            raise WMLClientError("Failed to start the file to  volume. Try again.")

    def upload_file(self, name,  file_path):
        """
            Upload the data file into stored volume.

            **Parameters**

                .. important::

                #. **name**:  Unique name of the stored volume \n
                   **type**: str\n

                #. **file_path**:  File to be uploaded into the volume \n
                   **type**: str\n

            **Output**
                .. important::

                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

                **Example**

                    >>> client.volumes.upload_file('testA', 'DRUG.csv')

        """
        WMLResource._chk_and_block_create_update_for_python36(self)

        if not self._client.ICP_PLATFORM_SPACES and not self._client.CLOUD_PLATFORM_SPACES:
            raise WMLClientError("Volume APIs are not supported. It is supported only for CP4D 3.5 and above.")

        header_input = self._client._get_headers(zen=True)
        zen_token = header_input.get('Authorization')

        filename_to_upload = file_path.split('/')[-1]
        upload_url_file = self._client.service_instance._href_definitions.volume_upload_href(name) + filename_to_upload
        cmd_str = 'curl -k  -X PUT "' + upload_url_file + '"' + "  -H 'Content-Type: multipart/form-data' -H 'Authorization: " + zen_token + \
                  "' -F upFile='@" + file_path + "'"
        args = shlex.split(cmd_str)
        upload_response = subprocess.run(args, capture_output=True, text=True)
        if upload_response.returncode == 0:
            import json
            try:
                cmd_output = json.loads(upload_response.stdout)
                print(cmd_output.get('message'))
                return "SUCCESS"
            except Exception as e:
                print(upload_response.returncode, upload_response.stdout)
                print("Failed to upload the file to  volume. Try again.")
                return "FAILED"
        else:
            print(upload_response.returncode, upload_response.stdout, upload_response.stderr)
            print("Failed to upload the file to  volume. Try again.")
            return "FAILED"

       # header = {'Authorization': zentoken}
       #  dirfiles = {'file': None}
       #  if not self._ICP:
       #      upload_dir_response = requests.post(
       #          upload_url,
       #          headers=header,
       #          files=dirfiles
       #          )
       #  else:
       #      upload_dir_response = requests.put(
       #          upload_url,
       #          headers=header,
       #          files=dirfiles,
       #          verify=False
       #      )
       #  if upload_dir_response.status_code != 200:
       #      raise WMLClientError("Failed to create a directory to upload the file to  volume. Try again.")
       #
       #
       #  filename_to_upload = file_path.split('/')[-1]
       #  #upload_url_file = self._client.service_instance._href_definitions.volume_upload_href(name) +
       #  upload_url_file = self._client.service_instance._href_definitions.volume_upload_href(name) + "wmldir/"
       # #  #upload_url_file = self._client.service_instance._href_definitions.volume_upload_href(name) +
       # #  import urllib
       # #  upload_url_file = self._client.service_instance._href_definitions.volume_upload_href(name)  + urllib.parse.quote(filename_to_upload, safe='')
       # #  print(upload_url_file)
       #  import urllib
       #  upload_url_file = self._client.service_instance._href_definitions.volume_upload_href(name) + filename_to_upload
       #  #upload_url_file = urllib.parse.quote(self._client.service_instance._href_definitions.volume_upload_href(name) + "wmldir/" +filename_to_upload, safe='')
       #  files = {'upload_file': (filename_to_upload,  open(file_path, 'rb'))}
       #
       #  with open(file_path, 'rb') as f:
       #      if not self._ICP:
       #          upload_response = requests.put(
       #              upload_url_file,
       #              headers=header,
       #              files=files
       #              )
       #      else:
       #          f = {'upload_file': (filename_to_upload, open(file_path, 'rb'), 'csv', {'Expires': '0'})}
       #          # with open(file_path, 'rb') as f:
       #          #     fdata = f.read()
       #              # upload_response = requests.put(
       #              #     upload_url_file,
       #              #     files={'file': (filename_to_upload, fdata, 'text/csv', {'Expires': '0'})},
       #              #     #files=files,
       #              #     #files=f,
       #              #     headers=header,
       #              #     verify=False
       #              # )
       #          # header.update({'Accept': 'text/csv'})#',image/png,image/gif',
       #          # upload_response = requests.request("PUT", upload_url_file,
       #          #                  files=f,
       #          #                  headers=header, verify=False)

    def list(self):
        """
           List stored scripts. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all script in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.volumes.list()
        """

        href = self._client.service_instance._href_definitions.volumes_href()
        params = {}
        params.update({'addon_type': 'volumes'})
        if not self._ICP:
            response = requests.get(href, params=params, headers=self._client._get_headers(zen=True))
        else:
            response = requests.get(href, params=params, headers=self._client._get_headers(zen=True), verify=False)
        asset_details = self._handle_response(200, u'list volumes', response)
        asset_list = asset_details.get('service_instances')
        volume_values = [
            (m[u'display_name'],
             m[u'id'],
             m['provision_status']) for m in asset_list]

        self._list(volume_values, [u'NAME', u'ID', u'PROVISION_STATUS'], None, _DEFAULT_LIST_LENGTH)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id(volume_details):
        """
                Get Unique Id of stored volume details.

                **Parameters**

                .. important::

                   #. **asset_details**:  Metadata of the stored volume details\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of stored volume asset\n
                    **return type**: str\n

                **Example**

                     >>> volume_uid = client.volumes.get_id(volume_details)

        """

        Volume._validate_type(volume_details, u'asset_details', object, True)
        if 'service_instance' in volume_details and  volume_details.get('service_instance') is not None:
            vol_details = volume_details.get('service_instance')
            return WMLResource._get_required_element_from_dict(vol_details, u'volume_assets_details',
                                                               [u'id'])
        else:
            return WMLResource._get_required_element_from_dict(volume_details, u'volume_assets_details',
                                                           [u'id'])

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_name(volume_details):
        """
                Get Unique name  of stored volume asset.

                **Parameters**

                .. important::

                   #. **volume_details**:  Metadata of the stored volume asset\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique name of stored volume asset\n
                    **return type**: str\n

                **Example**

                 >>> volume_name = client.volumes.get_name(asset_details)

        """
        Volume._validate_type(volume_details, u'asset_details', object, True)
        if 'service_instance' in volume_details and  volume_details.get('service_instance') is not None:
            vol_details = volume_details.get('service_instance')
            return WMLResource._get_required_element_from_dict(vol_details, u'volume_assets_details',
                                                               [u'display_name'])
        else:
            return WMLResource._get_required_element_from_dict(volume_details, u'volume_assets_details',
                                                           [u'display_name'])

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, volume_id):
        """
            Delete a volume.

            **Parameters**

            .. important::
                #. **volume_id**:  Unique ID of the volume.\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.volumes.delete(volume_id)

        """
        Volume._validate_type(volume_id, u'asset_uid', STR_TYPE, True)
        if (not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES):
            raise WMLClientError(u'Volume API is not supported.')

        if not self._ICP:
            response = requests.delete(self._client.service_instance._href_definitions.volume_href(volume_id),
                                       headers=self._client._get_headers(zen=True))
        else:
            response = requests.delete(self._client.service_instance._href_definitions.volume_href(volume_id),
                                      headers=self._client._get_headers(zen=True), verify=False)

        if response.status_code == 200 or response.status_code == 204:
            print("Successfully deleted volume service.")
            return "SUCCESS"
        else:
            print("Failed to delete volume.")
            print(response.status_code, response.text)
            return "FAILED"

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def stop(self, volume_name):
        """
            Stop the  volume service.

            **Parameters**

            .. important::
                #. **volume_name**:  Unique name of the volume
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.volumes.stop(volume_name)

        """
        Volume._validate_type(volume_name, u'asset_uid', STR_TYPE, True)
        if (not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES):
            raise WMLClientError(u'Volume API is not supported.')

        if not self._ICP:
            response = requests.delete(self._client.service_instance._href_definitions.volume_service_href(volume_name),
                                       headers=self._client._get_headers(zen=True))
        else:
            response = requests.delete(self._client.service_instance._href_definitions.volume_service_href(volume_name),
                                       headers=self._client._get_headers(zen=True), verify=False)
        if response.status_code == 200:
            print("Successfully stopped volume service.")
            return "SUCCESS"
        else:
            print(response.status_code, response.text)
            return "FAILED"




#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import logging
from ibm_watson_machine_learning.spaces import Spaces
from ibm_watson_machine_learning.tests.ICP.preparation_and_cleaning import *


class TestWMLClientWithSpace(unittest.TestCase):
    space_uid = None
    space_href=None
    member_uid = None
    member_href = None

    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithSpace.logger.info("Service Instance: setting up credentials")
        self.wml_credentials = get_wml_credentials()
        # reload(site)
        self.client = get_client()
        version = self.client.version


    # def test_01_service_instance_details(self):
    #     TestWMLClientWithSpace.logger.info("Check client ...")
    #     self.assertTrue(self.client.__class__.__name__ == 'APIClient')
    #
    #     TestWMLClientWithSpace.logger.info("Getting instance details ...")
    #     details = self.client.service_instance.get_details()
    #     TestWMLClientWithSpace.logger.debug(details)
    #
    #     self.assertTrue("published_models" in str(details))
    #     self.assertEqual(type(details), dict)

    #create pipeline first
    def test_01_save_space(self):
        metadata = {
                    self.client.repository.SpacesMetaNames.NAME: "V4Space"
                }

        space_details = self.client.repository.store_space(meta_props=metadata)

        TestWMLClientWithSpace.space_uid = self.client.repository.get_space_uid(space_details)
        TestWMLClientWithSpace.space_href = self.client.repository.get_space_href(space_details)

        space_specific_details = self.client.repository.get_space_details(TestWMLClientWithSpace.space_uid)
        self.assertTrue(TestWMLClientWithSpace.space_uid in str(space_specific_details))
        print(self.client.version)

    def test_02_set_space(self):
        set_space = self.client.set.default_space(TestWMLClientWithSpace.space_uid)


        self.assertTrue("SUCCESS" in str(set_space))

    def test_03_update_space(self):
        metadata = {
            self.client.repository.SpacesMetaNames.NAME: "my_space_TestWMLClientWithSpace",
            self.client.repository.SpacesMetaNames.DESCRIPTION: "mnist best model",
        }

        space_details = self.client.repository.update_space(TestWMLClientWithSpace.space_uid, metadata)
        self.assertTrue('my_space_TestWMLClientWithSpace' in str(space_details))
        TestWMLClientWithSpace.logger.info(space_details)
        self.assertTrue('V4Space' not in str(space_details))

    def test_04_get_space_details(self):
        details = self.client.repository.get_space_details()
        self.assertTrue(TestWMLClientWithSpace.space_uid in str(details))

        details2 = self.client.repository.get_space_details(TestWMLClientWithSpace.space_uid)
        self.assertTrue(TestWMLClientWithSpace.space_uid in str(details2))

    def test_05_list(self):
        stdout_ = sys.stdout
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.repository.list_spaces()  # Call function.
        sys.stdout = stdout_  # Reset redirect.
        self.assertTrue(TestWMLClientWithSpace.space_uid in captured_output.getvalue())
        self.client.repository.list_spaces()  # Just to see values.

    def test_06_create_member(self):
        metadata = {
            self.client.spaces.MemberMetaNames.ROLE: "Viewer",
            self.client.spaces.MemberMetaNames.IDENTITY_TYPE: "service",
            self.client.spaces.MemberMetaNames.IDENTITY: "IBMid-310002RQJW"
        }

        member_details = self.client.spaces.create_member(space_uid=TestWMLClientWithSpace.space_uid,meta_props=metadata)

        TestWMLClientWithSpace.member_uid = self.client.spaces.get_member_uid(member_details)
        TestWMLClientWithSpace.member_href = self.client.spaces.get_member_href(member_details)

        member_specific_details = self.client.spaces.get_members_details(TestWMLClientWithSpace.space_uid,TestWMLClientWithSpace.member_uid)
        self.assertTrue(TestWMLClientWithSpace.member_uid in str(member_specific_details))

    def test_07_get_member_details(self):
        member_specific_details = self.client.spaces.get_members_details(TestWMLClientWithSpace.space_uid,TestWMLClientWithSpace.member_uid)
        self.assertTrue(TestWMLClientWithSpace.member_uid in str(member_specific_details))

    def test_08_delete_member(self):
        delete_member = self.client.spaces.delete_members(TestWMLClientWithSpace.space_uid,TestWMLClientWithSpace.member_uid)
        self.assertTrue("SUCCESS" in str(delete_member))
    def test_09_delete_space(self):
        delete_space = self.client.repository.delete(TestWMLClientWithSpace.space_uid)
        self.assertTrue("SUCCESS" in str(delete_space))




__all__ = [
    'FlightService'
]

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import json
import sys
import threading
from typing import List, Optional, Iterable, TYPE_CHECKING
from .errors import DataStreamError, WrongLocationProperty, FlightServiceInfoNotAvailable, WrongFileLocation, \
    FileUploadFailed
from .utils import prepare_interaction_props_for_cos

import pandas as pd

if TYPE_CHECKING:
    from ibm_watson_machine_learning import APIClient

# the upper bound limitation for training data size
DATA_SIZE_LIMIT = 1073741824  # 1GB in Bytes


class FlightService:
    """FlightService object unify the work for training data reading from different types of data sources,
        including databases. It uses a Flight Service and `pyarrow` library to connect and transfer the data.

    Parameters
    ----------
    wml_client: APIClient, required
        WML Client

    n_batches: int, optional
        Defines for how many parts / batches data source should be split.
        Default is 1.

    data_location: dict, optional
        Data location information passed by user.

    """

    def __init__(self,
                 wml_client: 'APIClient',
                 params: dict,
                 n_batches: Optional[int] = 1,
                 data_location: dict = None,
                 experiment_metadata: dict = None) -> None:
        self.wml_client = wml_client
        self.params = params
        self.n_batches = n_batches
        self.data_source_type = None
        self.experiment_metadata = experiment_metadata

        self.data_location = data_location

        self.lock_read = threading.Lock()
        self.stop_reading = False

        from pyarrow import flight

        flight_hostname = os.environ.get('FLIGHT_SERVICE_LOCATION', 'wdp-connect-flight')
        flight_port = os.environ.get('FLIGHT_SERVICE_PORT', '443')

        if flight_hostname == 'wdp-connect-flight':
            try:
                from project_lib import Project

            except ModuleNotFoundError:
                raise FlightServiceInfoNotAvailable()

        self.flight_client = flight.FlightClient(
            location=f"grpc+tls://{flight_hostname}:{flight_port}",
            disable_server_verification=True,
            override_hostname=flight_hostname
        )

    def authenticate(self) -> 'flight.ClientAuthHandler':
        """Creates an authenticator object for Flight Service."""
        from pyarrow import flight

        class TokenClientAuthHandler(flight.ClientAuthHandler):
            """Authenticator implementation from pyarrow flight."""

            def __init__(self, token):
                super().__init__()
                self.token = bytes('Bearer ' + token, 'utf-8')

            def authenticate(self, outgoing, incoming):
                outgoing.write(self.token)
                self.token = incoming.read()

            def get_token(self):
                return self.token

        if self.wml_client is None:
            return TokenClientAuthHandler(token=os.environ.get('USER_ACCESS_TOKEN'))
        else:
            return TokenClientAuthHandler(token=self.wml_client.wml_token)

    def get_endpoints(self) -> Iterable[List['flight.FlightEndpoint']]:
        """Listing all available Flight Service endpoints (one endpoint corresponds to one batch)"""
        from pyarrow import flight

        self.flight_client.authenticate(self.authenticate())
        try:

            for source_command in self._select_source_command():
                info = self.flight_client.get_flight_info(
                    flight.FlightDescriptor.for_command(source_command)
                )
                yield info.endpoints

        except flight.FlightInternalError as e:
            if 'CDICO2034E' in str(e):
                if 'The property [infer_schema] is not supported.' in str(e):
                    for source_command in self._select_source_command(infer_schema=False):
                        info = self.flight_client.get_flight_info(
                            flight.FlightDescriptor.for_command(source_command)
                        )
                        yield info.endpoints

                else:
                    raise WrongLocationProperty(e)

            elif 'CDICO2015E' in str(e):
                raise WrongFileLocation(e)

            else:
                raise e

    def _get_data(self, endpoint: 'flight.FlightEndpoint') -> 'pd.DataFrame':
        """
        Read data from Flight Service (only one batch).

        Properties
        ----------
        endpoint: flight.FlightEndpoint, required

        Returns
        -------
        pd.DataFrame with batch data
        """
        from pyarrow import flight
        import pyarrow as pa

        try:
            reader = self.flight_client.do_get(endpoint.ticket)

        except flight.FlightUnavailableError as e:
            raise DataStreamError(reason=str(e))

        chunks = []

        # Flight Service could split one batch into several chunks to have better performance
        while True:
            try:
                chunk, metadata = reader.read_chunk()
                chunks.append(chunk)
            except StopIteration:
                break

        if len(chunks) == 0:
            return None

        else:
            data = pa.Table.from_batches(chunks)
            return data.to_pandas()

    @staticmethod
    def _cast_columns_to_float_64(data: 'pd.DataFrame') -> 'pd.DataFrame':
        """Flight Service will cast decfloat types to strings, we need to cast them to correct types."""
        for i, (col_name, col_type) in enumerate(zip(data.dtypes.index, data.dtypes)):

            if col_type == 'object':
                print(col_name)

                try:
                    data[col_name] = data[col_name].astype('float64')

                except ValueError:  # ignore when column cannot be cast to other type than string
                    pass

                except Exception as e:
                    pass

        return data

    def _read_batch(self, batch_number: int, sequence_number: int,
                    endpoint: 'flight.FlightEndpoint', dfs: List['pd.DataFrame']) -> None:
        """This method should be used as a separate thread for downloading specific batch of data in parallel.

        Parameters
        ----------
        batch_number: int, required
            Specific number of the batch to download.

        sequence_number: int required
            Sequence number, it tells the number of file that we reads for file storage.

        endpoint: flight.FlightEndpoint, required
            Flight Service endpoint to connect.

        dfs: List['pd.DataFrame'], required
            List where we will be storing the downloaded batch. Shared across threads.
        """
        if not self.stop_reading:
            df = self._get_data(endpoint)
            if df is not None:
                row_size = sys.getsizeof(df.iloc[:2]) - sys.getsizeof(df.iloc[:1])

                if row_size != 0:  # append batches only when we have data
                    with self.lock_read:
                        if not self.stop_reading:
                            # note: what to do if we have batch too large (over 1 GB limit)
                            if sys.getsizeof(df) > DATA_SIZE_LIMIT:  # this is GB in Bytes
                                row_limit = round(DATA_SIZE_LIMIT / row_size)
                                dfs.insert(batch_number, df.iloc[:row_limit])
                                self.stop_reading = True
                                # print(f"STOP READING SET by batch: {batch_number}")
                            # --- end note
                            else:
                                total_size = sum([row_size * len(data) for data in dfs])

                                # note: what to do when we have total size nearly under the limit
                                if total_size <= DATA_SIZE_LIMIT:
                                    upper_row_limit = (DATA_SIZE_LIMIT - total_size) // row_size
                                    df = df.iloc[:upper_row_limit]
                                    dfs.insert(batch_number, df)
                                # --- end note

                                else:
                                    # print(f"total size stop: {total_size}")
                                    self.stop_reading = True

            # print(f"Downloaded batch number: {batch_number}, sequence: {sequence_number}")
            # print(f"Batch shape: {df.shape}")
            # print(f"Estimated batch size: {sys.getsizeof(df)} Bytes")
            # print(f'Thread {batch_number}, sequence: {sequence_number} completed reading the batch.')

    def read(self) -> 'pd.DataFrame':
        """Fetch the data from Flight Service. Fetching is done in batches.
            There is an upper top limit of data size to be fetched configured to 1 GB.

        Returns
        -------
        Pandas DataFrame with fetched data.
        """
        dfs = []
        sequences = []

        # Note: endpoints are created by Flight Service based on number of partitions configured
        # one endpoint serves one batch of the data
        for n, endpoints in enumerate(self.get_endpoints()):
            threads = []
            sequences.append(threads)

            for i, endpoint in enumerate(endpoints):
                reading_thread = threading.Thread(target=self._read_batch, args=(i, n, endpoint, dfs))
                threads.append(reading_thread)
                # print(f"Starting batch reading thread: {i}, sequence: {n}...")
                reading_thread.start()

        for n, sequence in enumerate(sequences):
            for i, thread in enumerate(sequence):
                # print(f"Joining batch reading thread {i}, sequence: {n}...")
                thread.join()

        dfs = self._cast_columns_to_float_64(data=pd.concat(dfs))

        # Note: be sure that we do not cross upper data size limit
        estimated_data_size = sys.getsizeof(dfs)
        estimated_row_size = estimated_data_size // len(dfs)
        size_over_limit = estimated_data_size - DATA_SIZE_LIMIT

        if size_over_limit > 0:
            upper_limit = len(dfs) - (size_over_limit // estimated_row_size) * 2
            dfs = dfs.iloc[:upper_limit]
        # --- end note

        print(f"TOTAL DOWNLOADED DATA SIZE: {sys.getsizeof(dfs)} Bytes")

        return dfs

    def get_asset_id(self) -> dict:
        _dict = {
            "data_asset_id": None,
            "connection_asset_id": None,
        }

        if self.data_location['type'] == 'data_asset':
            _dict['data_asset_id'] = self.data_location['href'].split('/v2/assets/')[-1].split('?')[0]

        else:
            _dict['connection_asset_id'] = self.data_location['connection']['id']

        return _dict

    def _select_source_command(self, infer_schema: bool = True) -> List[str]:
        """Based on a data source type, select appropriate commands for flight service configuration."""
        # when dataset is small we received empty dfs, it is ok (15 is optimum for larger than 1 GB)
        command = {
            "num_partitions": 15,
        }

        # note: WS scenario
        if self.wml_client is None:
            project_id = os.environ.get('PROJECT_ID')
            if project_id is None:
                command['project_id'] = self.experiment_metadata.get('project_id')

            else:
                command['project_id'] = project_id

        else:
            if self.wml_client.default_space_id is not None:
                command['space_id'] = self.wml_client.default_space_id

            elif self.wml_client.default_project_id is not None:
                command['project_id'] = self.wml_client.default_project_id

        ids = self.get_asset_id()

        if ids['data_asset_id']:
            command['asset_id'] = ids['data_asset_id']
            if infer_schema:
                command['interaction_properties'] = {'infer_schema': "true"}

        elif ids['connection_asset_id']:
            command['asset_id'] = ids['connection_asset_id']
            command['interaction_properties'] = {}

            if 'bucket' in self.data_location['location']:
                command['interaction_properties'] = prepare_interaction_props_for_cos(
                    self.params, self.data_location['location'].get(
                        'path', self.data_location['location']['file_name']))
                command['interaction_properties']['infer_schema'] = "true"
                command['interaction_properties']['file_name'] = self.data_location['location'].get(
                    'path', self.data_location['location']['file_name'])
                command['interaction_properties']['bucket'] = self.data_location['location']['bucket']

            elif 'schema_name' in self.data_location['location']:
                command['interaction_properties']['schema_name'] = self.data_location['location']['schema_name']
                command['interaction_properties']['table_name'] = self.data_location['location']['table_name']

            else:
                command['interaction_properties'] = self.data_location['location']
                command['interaction_properties']['infer_schema'] = "true"

        return [json.dumps(command)]

    def write_data(self, data: 'pd.DataFrame'):
        from pyarrow import flight
        import pyarrow as pa

        schema = pa.Schema.from_pandas(data)
        ids = self.get_asset_id()
        command = {}

        # note: WS scenario
        if self.wml_client is None:
            command['project_id'] = os.environ.get('PROJECT_ID')

        else:
            if self.wml_client.default_space_id is not None:
                command['space_id'] = self.wml_client.default_space_id

            elif self.wml_client.default_project_id is not None:
                command['project_id'] = self.wml_client.default_project_id

        if ids['data_asset_id']:
            command['asset_id'] = ids['data_asset_id']

        elif ids['connection_asset_id']:
            command['asset_id'] = ids['connection_asset_id']
            command['interaction_properties'] = {}

            if 'bucket' in self.data_location['location']:
                command['interaction_properties']['file_name'] = self.data_location['location'].get(
                    'path', self.data_location['location']['file_name'])
                command['interaction_properties']['bucket'] = self.data_location['location']['bucket']

            elif 'schema_name' in self.data_location['location']:
                command['interaction_properties']['schema_name'] = self.data_location['location']['schema_name']
                command['interaction_properties']['table_name'] = self.data_location['location']['table_name']

            else:
                command['interaction_properties'] = self.data_location['location']

        self.flight_client.authenticate(self.authenticate())
        writer, reader = self.flight_client.do_put(flight.FlightDescriptor.for_command(json.dumps(command)), schema)

        with writer:
            writer.write_table(pa.Table.from_pandas(data))

        return writer, reader



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import inspect
from abc import ABCMeta
from py4j.java_collections import JavaArray, JavaList
from py4j.java_collections import ListConverter
from py4j.java_gateway import JavaObject
from py4j.protocol import Py4JJavaError
from pyspark import RDD, SparkContext
from pyspark.ml.param import Params
from pyspark.ml.wrapper import JavaModel
from pyspark.serializers import PickleSerializer, AutoBatchedSerializer
from pyspark.sql import DataFrame, SQLContext



class PythonJavaConversions(object):

    _sc = SparkContext._active_spark_context
    _logSrcLang = "Py:"

    @classmethod
    def _resetUid(cls, stage, uid):

        logger = cls._sc._jvm.org.apache.log4j.Logger.getLogger(cls.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = cls._logSrcLang + cls.__name__ + ":" + methodname + ": [Params: " + "stage => " + str(
                stage) + "\" | uid => \"" + str(uid) + "\"]"
        logger.info(logMsg)

        newUid = unicode(uid)
        stage.uid = newUid
        newDefaultParamMap = dict()
        newParamMap = dict()
        for param in stage.params:
            newParam = copy.copy(param)
            newParam.parent = newUid
            if param in stage._defaultParamMap:
                newDefaultParamMap[newParam] = stage._defaultParamMap[param]
            if param in stage._paramMap:
                newParamMap[newParam] = stage._paramMap[param]
            param.parent = newUid
        stage._defaultParamMap = newDefaultParamMap
        stage._paramMap = newParamMap
        return stage

    @classmethod
    def _to_java_stage(cls, stage):
        logger = cls._sc._jvm.org.apache.log4j.Logger.getLogger(cls.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = cls._logSrcLang + cls.__name__ + ":" + methodname + ": [Params: " + "stage => " + str(stage) + "]"
        logger.info(logMsg)

        paramMap = stage.extractParamMap()
        for param in stage.params:
            if param in paramMap:
                pair = cls._make_java_param_pair(stage, param, paramMap[param])
                stage._java_obj.set(pair)
        return stage._java_obj

    @classmethod
    def _py2java(cls, sc, obj):
        logger = cls._sc._jvm.org.apache.log4j.Logger.getLogger(cls.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = cls._logSrcLang + cls.__name__ + ":" + methodname + ": [Params: " + "sc => " + str(
                sc) + " | obj => " + str(obj) + "]"
        logger.info(logMsg)

        """ Convert Python object into Java """
        if isinstance(obj, RDD):
            obj = cls._to_java_object_rdd(obj)
        elif isinstance(obj, DataFrame):
            obj = obj._jdf
        elif isinstance(obj, SparkContext):
            obj = obj._jsc
        elif isinstance(obj, list):
            obj = ListConverter().convert([cls._py2java(sc, x) for x in obj],
                                          sc._gateway._gateway_client)
        elif isinstance(obj, JavaObject):
            pass
        elif isinstance(obj, (int, long, float, bool, bytes, unicode)):
            pass
        else:
            data = bytearray(PickleSerializer().dumps(obj))
            obj = sc._jvm.SerDe.loads(data)
        return obj

    """
    Krishna: _java2py is called only by _call_java(). But _call_java is not used anywhere. Hence commenting it for now

    @classmethod
    def _java2py(sc, r, encoding="bytes"):
        _logger = _jLogger.getLogger(cls.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = _logSrcLang + cls.__name__ + ":" + methodname + ": [Params: " + "sc => " + str(sc) + " | r => " + str(
                r) + " | encoding => " + str(encoding) + "]"
        _logger.info(logMsg)

        if isinstance(r, JavaObject):
            clsName = r.getClass().getSimpleName()
            # convert RDD into JavaRDD
            if clsName != 'JavaRDD' and clsName.endswith("RDD"):
                r = r.toJavaRDD()
                clsName = 'JavaRDD'

            if clsName == 'JavaRDD':
                jrdd = sc._jvm.SerDe.javaToPython(r)
                return RDD(jrdd, sc)

            if clsName == 'Dataset':
                return DataFrame(r, SQLContext.getOrCreate(sc))

            if clsName in _picklable_classes:
                r = sc._jvm.SerDe.dumps(r)
            elif isinstance(r, (JavaArray, JavaList)):
                try:
                    r = sc._jvm.SerDe.dumps(r)
                except Py4JJavaError:
                    pass  # not pickable

        if isinstance(r, (bytearray, bytes)):
            r = PickleSerializer().loads(bytes(r), encoding=encoding)
            return r
    """

    @staticmethod
    def _to_java_object_rdd(rdd):
        """ Return an JavaRDD of Object by unpickling

        It will convert each Python object into Java object by Pyrolite, whenever the
        RDD is serialized in batch or not.
        """
        rdd = rdd._reserialize(AutoBatchedSerializer(PickleSerializer()))
        return rdd.ctx._jvm.SerDe.pythonToJava(rdd._jrdd, True)

    @classmethod
    def _make_java_param_pair(cls, stage, param, value):
        logger = cls._sc._jvm.org.apache.log4j.Logger.getLogger(cls.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = cls._logSrcLang + cls.__name__ + ":" + methodname + ": [Params: " + "stage => " + str(
                stage) + " | param => " + str(param) + " | encoding => " + str(value) + "]"
        logger.info(logMsg)

        param = stage._resolveParam(param)
        java_param = stage._java_obj.getParam(param.name)
        java_value = cls._py2java(cls._sc, value)
        return java_param.w(java_value)

    @classmethod
    def _from_java_stage(cls, java_stage):
        """
        Given a Java object, create and return a Python wrapper of it.
        Used for ML persistence.

        Meta-algorithms such as Pipeline should override this method as a classmethod.
        """

        def __get_class(clazz):
            """
            Loads Python class from its name.
            """
            parts = clazz.split('.')
            module = ".".join(parts[:-1])
            m = __import__(module)
            for comp in parts[1:]:
                m = getattr(m, comp)
            return m

        logger = cls._sc._jvm.org.apache.log4j.Logger.getLogger(cls.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = cls._logSrcLang + cls.__name__ + ":" + methodname + ": [Params: " + "java_stage => " + str(
                java_stage) + "]"
        logger.info(logMsg)

        stage_name = java_stage.getClass().getName().replace("org.apache.spark", "pyspark")
        # Generate a default new instance from the stage_name class.
        py_type = __get_class(stage_name)

        if issubclass(py_type, JavaModel):
            py_stage = py_type(java_stage)
        else:
            py_stage = py_type()

        # Load information from java_stage to the instance.
        py_stage._java_obj = java_stage
        py_stage = cls._resetUid(py_stage, java_stage.uid())
        py_stage._transfer_params_from_java()
        return py_stage

    @staticmethod
    def _to_java_list(java_obj, scala_seq):
        return java_obj.toJavaList(scala_seq)

    @classmethod
    def _load_java_obj(cls, clazz):
        """
        >>> Load the peer Java object of the ML instance.
        """
        # java_class = cls._java_loader_class(clazz)

        logger = cls._sc._jvm.org.apache.log4j.Logger.getLogger(cls.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = cls._logSrcLang + cls.__name__ + ":" + methodname + ": [Params: " + "clazz => " + str(clazz) + "]"
        logger.info(logMsg)

        if clazz.__name__ == "IBMSparkPipeline":
            java_class = "com.ibm.analytics.ngp.pipeline.IBMSparkPipeline"

        else:
            java_class = "org.apache.spark.ml.IBMSparkPipelineModel"

        java_obj = cls._sc._jvm
        for name in java_class.split("."):
            java_obj = getattr(java_obj, name)

        return java_obj


class JavaWrapper(object):
    """
    Wrapper class for a Java companion object
    """
    _sc = SparkContext._active_spark_context
    _logSrcLang = "Py:"

    def __init__(self, java_obj=None):
        super(JavaWrapper, self).__init__()
        self._java_obj = java_obj


    """
    Krishna: _create_from_java_class is not used anywhere. Hence commenting it for now
    @classmethod
    def _create_from_java_class(cls, java_class, *args):

        # Construct this object from given Java classname and arguments

        _logger = _jLogger.getLogger(cls.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = _logSrcLang + cls.__name__ + ":" + methodname + ": [Params: " + "java_class => " + str(
                java_class) + " | args => "
        for curArg in args[:-1]:
            logMsg = logMsg + str(curArg) + ","
        logMsg = logMsg + args[-1] + "]"
        _logger.info(logMsg)

        java_obj = JavaWrapper._new_java_obj(java_class, *args)
        return cls(java_obj)
    """

    """
    Krishna: _call_java is not used anywhere. Hence commenting it for now
    def _call_java(self, name, *args):
        m = getattr(self._java_obj, name)
        sc = SparkContext._active_spark_context
        java_args = [PythonJavaConversions._py2java(sc, arg) for arg in args]
        return PythonJavaConversions._java2py(sc, m(*java_args))
    """

    @classmethod
    def _new_java_obj(cls,java_class, *args):
        """
        Returns a new Java object.
        """
        logger = cls._sc._jvm.org.apache.log4j.Logger.getLogger(cls.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = cls._logSrcLang + 'JavaWrapper' + ":" + methodname + ": [Params: " + "java_class => " + str(
                java_class) + " | args => "
        for curArg in args[:-1]:
            logMsg = logMsg + str(curArg) + ","
        logMsg = logMsg + str(args[-1]) + "]"
        logger.info(logMsg)

        sc = SparkContext._active_spark_context
        java_obj = SparkContext._active_spark_context._jvm
        for name in java_class.split("."):
            java_obj = getattr(java_obj, name)
        java_args = [PythonJavaConversions._py2java(sc, arg) for arg in args]
        return java_obj(*java_args)


class JavaParams(JavaWrapper, Params):
    """
    Utility class to help create wrapper classes from Java/Scala
    implementations of pipeline components.
    """
    #: The param values in the Java object should be
    #: synced with the Python wrapper in fit/transform/evaluate/copy.

    __metaclass__ = ABCMeta



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import traceback

from lale.operators import BasePipeline

from ibm_watson_machine_learning.experiment.autoai.optimizers import RemoteAutoPipelines

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType

test_case = unittest.TestCase()


def is_lale_pipeline_type(pipeline: object) -> None:
    print("Testing if object is of lale pipeline type")
    test_case.assertIsInstance(pipeline, BasePipeline)


def validate_autoai_experiment(experiment_info, autoai_pod_version, data_join_graph = None):
    """
    Checks if all requiered parameters for AutoAI experiment was set.
    Also adds autoai pod version if missing.

    Returns
    -------
    Dictionary with checked experiment info.
    """

    required_params = ['name', 'prediction_type', 'prediction_column']

    for param in required_params:
        test_case.assertIn(param, experiment_info.keys(),
                           msg=f"{param} is missing in experiment parameters {experiment_info}")

    if "autoai_pod_version" not in experiment_info.keys():
        experiment_info['autoai_pod_version'] = autoai_pod_version

    if data_join_graph:
        experiment_info['data_join_graph'] = data_join_graph

    return experiment_info


def validate_autoai_timeseries_experiment(experiment_info, autoai_pod_version):
    """
    Checks if all requiered parameters for AutoAI FORECASTING experiment was set.
    Also adds autoai pod version if missing.

    Returns
    -------
    Dictionary with checked experiment info.
    """

    required_params = ['name', 'prediction_columns', 'timestamp_column_name']

    for param in required_params:
        test_case.assertIn(param, experiment_info.keys(),
                           msg=f"{param} is missing in experiment parameters {experiment_info}")

    if "prediction_type" not in experiment_info.keys():
        experiment_info['prediction_type'] = PredictionType.FORECASTING

    test_case.assertEqual(experiment_info['prediction_type'], PredictionType.FORECASTING,
                          msg="Prediction type for forecasting experiment was set incorrectly ")

    if "autoai_pod_version" not in experiment_info.keys():
        experiment_info['autoai_pod_version'] = autoai_pod_version

    return experiment_info


def get_and_predict_all_pipelines_as_lale(remote_auto_pipelines: 'RemoteAutoPipelines', x_values):
    """
    Based on summary of the experiment the method is loading each pipeline to lale type and check if predict is working.
    """

    from lale.operators import TrainablePipeline

    summary = remote_auto_pipelines.summary()
    print(summary)

    failed_pipelines = []
    error_messages = []

    for pipeline_name in summary.reset_index()['Pipeline Name']:
        print(f"Getting pipeline: {pipeline_name}")
        try:
            lale_pipeline = remote_auto_pipelines.get_pipeline(pipeline_name=pipeline_name)
            test_case.assertIsInstance(lale_pipeline, TrainablePipeline)
            predictions = lale_pipeline.predict(X=x_values[:1])
            print(predictions)
            test_case.assertGreater(len(predictions), 0, msg=f"Returned prediction for {pipeline_name} are empty")
        except:
            print(f"Failure: {pipeline_name}")
            failed_pipelines.append(pipeline_name)
            error_message = traceback.format_exc()
            print(error_message)
            error_messages.append(error_message)

    test_case.assertEqual(len(failed_pipelines), 0, msg=f"Some pipelines failed. Full list: {failed_pipelines} \n "
                                                       f"Errors: {error_messages}")



import unittest
from os import environ
import requests
import json

import ibm_boto3

from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.experiment.autoai.optimizers import RemoteAutoPipelines
from ibm_watson_machine_learning.helpers.connections import DataConnection, ContainerLocation
from ibm_watson_machine_learning.tests.utils import is_cp4d, save_data_to_container, get_wml_credentials, \
    get_cos_credentials, get_space_id
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestWebservice, \
    AbstractTestAutoAIAsync, AbstractTestBatch
from ibm_watson_machine_learning.tests.utils.assertions import validate_autoai_experiment
from ibm_watson_machine_learning.tests.utils.cleanup import space_cleanup

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics, ClassificationAlgorithms, \
    ImputationStrategy
from ibm_watson_machine_learning.utils.autoai.errors import StrategyIsNotApplicable, InvalidImputationParameterNonTS, \
    InvalidImputationParameterTS, NumericalImputationStrategyValueMisused


class TestAutoAIRemote(unittest.TestCase):
    """
    The test can be run on CLOUD, and CPD
    """

    cos_resource = None
    data_location = './autoai/data/Corona_NLP_test_utf8.csv'

    data_cos_path = 'data/Corona_NLP_test_utf8.csv'

    SPACE_ONLY = False

    OPTIMIZER_NAME = "Corona NLP Text Transformer test sdk"

    batch_payload_location = './autoai/data/Corona_NLP_scoring_payload.csv'
    batch_payload_cos_location = "Corona_NLP_scoring_payload.csv"

    BATCH_DEPLOYMENT_WITH_CA = True
    BATCH_DEPLOYMENT_WITH_CDA = False
    BATCH_DEPLOYMENT_WITH_DA = False
    BATCH_DEPLOYMENT_WITH_DF = True

    target_space_id = None
    pod_version = environ.get('KB_VERSION', None)

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        prediction_type=PredictionType.MULTICLASS,
        prediction_column='Sentiment',
        scoring=Metrics.F1_SCORE_MACRO,
        holdout_size=0.1,
        text_processing=True,
        text_columns_names=['OriginalTweet', 'Location'],
        word2vec_feature_number=4,
        max_number_of_estimators=1,
        daub_give_priority_to_runtime=3.0,
    )

    ts_experiment_info = dict(
        name=OPTIMIZER_NAME,
        desc='test description',
        prediction_type=PredictionType.FORECASTING,
        prediction_columns=['value1', 'value2'],
        timestamp_column_name='timestamp',
        backtest_num=4,
        lookback_window=5,
        forecast_window=2,
        holdout_size=0.05,
        max_number_of_estimators=1,
        include_only_estimators=[AutoAI.ForecastingAlgorithms.ENSEMBLER],
        #csv_separator=','
    )

    space_name = environ.get('SPACE_NAME', 'regression_tests_sdk_space')
    space_id = None
    experiment = None

    @classmethod
    def setUpClass(cls) -> None:
        """
        Load WML credentials from config.ini file based on ENV variable.
        """
        cls.wml_credentials = get_wml_credentials()
        cls.wml_client = APIClient(wml_credentials=cls.wml_credentials.copy())

        cls.cos_credentials = get_cos_credentials()
        cls.cos_endpoint = cls.cos_credentials.get('endpoint_url')
        cls.cos_resource_instance_id = cls.cos_credentials.get('resource_instance_id')

        cls.project_id = cls.wml_credentials.get('project_id')

    def test_00a_space_cleanup(self):
        space_checked = False
        while not space_checked:
            space_cleanup(self.wml_client,
                          get_space_id(self.wml_client, self.space_name,
                                       cos_resource_instance_id=self.cos_resource_instance_id),
                          days_old=7)
            space_id = get_space_id(self.wml_client, self.space_name,
                                    cos_resource_instance_id=self.cos_resource_instance_id)
            try:
                self.assertIsNotNone(space_id, msg="space_id is None")
                space_checked = True
            except AssertionError:
                space_checked = False

        TestAutoAIRemote.space_id = space_id

        if self.SPACE_ONLY:
            self.wml_client.set.default_space(self.space_id)
        else:
            self.wml_client.set.default_project(self.project_id)

    def test_01_initialize_AutoAI_experiment__pass_credentials__object_initialized(self):
        if self.SPACE_ONLY:
            TestAutoAIRemote.experiment = AutoAI(wml_credentials=self.wml_credentials.copy(),
                                                        space_id=self.space_id)
        else:
            TestAutoAIRemote.experiment = AutoAI(wml_credentials=self.wml_credentials.copy(),
                                                        project_id=self.project_id)

        self.assertIsInstance(self.experiment, AutoAI, msg="Experiment is not of type AutoAI.")

    def test_02_initialize_optimizer_kb_scenario(self):
        experiment_info = self.experiment_info.copy()
        experiment_info = validate_autoai_experiment(experiment_info, self.pod_version)

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0]\
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertNotIn('use_imputation', params)
        self.assertNotIn('imputer_list', params)
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_03_initialize_optimizer_kb_scenario_mean(self):
        experiment_info = self.experiment_info.copy()
        experiment_info['categorical_imputation_strategy'] = ImputationStrategy.MEAN
        experiment_info = validate_autoai_experiment(experiment_info, self.pod_version)

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertIn('preprocessor_cat_imp_strategy', params)
        self.assertEqual(params['preprocessor_cat_imp_strategy'], 'mean')
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertNotIn('use_imputation', params)
        self.assertNotIn('imputer_list', params)
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_04_initialize_optimizer_kb_scenario_median(self):
        experiment_info = self.experiment_info.copy()
        experiment_info['categorical_imputation_strategy'] = ImputationStrategy.MEDIAN
        experiment_info = validate_autoai_experiment(experiment_info, self.pod_version)

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertIn('preprocessor_cat_imp_strategy', params)
        self.assertEqual(params['preprocessor_cat_imp_strategy'], 'median')
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertNotIn('use_imputation', params)
        self.assertNotIn('imputer_list', params)
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_05_initialize_optimizer_kb_scenario_most_frequent(self):
        experiment_info = self.experiment_info.copy()
        experiment_info['categorical_imputation_strategy'] = ImputationStrategy.MOST_FREQUENT
        experiment_info = validate_autoai_experiment(experiment_info, self.pod_version)

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertIn('preprocessor_cat_imp_strategy', params)
        self.assertEqual(params['preprocessor_cat_imp_strategy'], 'most_frequent')
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertNotIn('use_imputation', params)
        self.assertNotIn('imputer_list', params)
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_06_initialize_optimizer_kb_scenario_invalid_strategies(self):
        for strategy in [ImputationStrategy.BEST_OF_DEFAULT_IMPUTERS, ImputationStrategy.VALUE,
                         ImputationStrategy.FLATTEN_ITERATIVE, ImputationStrategy.LINEAR, ImputationStrategy.CUBIC,
                         ImputationStrategy.PREVIOUS, ImputationStrategy.NEXT, ImputationStrategy.NO_IMPUTATION]:
            with self.assertRaises(StrategyIsNotApplicable):
                experiment_info = self.experiment_info.copy()
                experiment_info['categorical_imputation_strategy'] = strategy
                experiment_info = validate_autoai_experiment(experiment_info, self.pod_version)

                TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**experiment_info)

    def test_07_initialize_optimizer_kb_scenario_mean_num(self):
        experiment_info = self.experiment_info.copy()
        experiment_info['numerical_imputation_strategy'] = ImputationStrategy.MEAN
        experiment_info = validate_autoai_experiment(experiment_info, self.pod_version)

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertIn('preprocessor_num_imp_strategy', params)
        self.assertEqual(params['preprocessor_num_imp_strategy'], 'mean')
        self.assertNotIn('use_imputation', params)
        self.assertNotIn('imputer_list', params)
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_08_initialize_optimizer_kb_scenario_median_num(self):
        experiment_info = self.experiment_info.copy()
        experiment_info['numerical_imputation_strategy'] = ImputationStrategy.MEDIAN
        experiment_info = validate_autoai_experiment(experiment_info, self.pod_version)

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertIn('preprocessor_num_imp_strategy', params)
        self.assertEqual(params['preprocessor_num_imp_strategy'], 'median')
        self.assertNotIn('use_imputation', params)
        self.assertNotIn('imputer_list', params)
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_09_initialize_optimizer_kb_scenario_most_frequent_num(self):
        experiment_info = self.experiment_info.copy()
        experiment_info['numerical_imputation_strategy'] = ImputationStrategy.MOST_FREQUENT
        experiment_info = validate_autoai_experiment(experiment_info, self.pod_version)

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertIn('preprocessor_num_imp_strategy', params)
        self.assertEqual(params['preprocessor_num_imp_strategy'], 'most_frequent')
        self.assertNotIn('use_imputation', params)
        self.assertNotIn('imputer_list', params)
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_10_initialize_optimizer_kb_scenario_invalid_strategies_num(self):
        for strategy in [ImputationStrategy.BEST_OF_DEFAULT_IMPUTERS, ImputationStrategy.VALUE,
                         ImputationStrategy.FLATTEN_ITERATIVE, ImputationStrategy.LINEAR, ImputationStrategy.CUBIC,
                         ImputationStrategy.PREVIOUS, ImputationStrategy.NEXT, ImputationStrategy.NO_IMPUTATION]:
            with self.assertRaises(StrategyIsNotApplicable):
                experiment_info = self.experiment_info.copy()
                experiment_info['numerical_imputation_strategy'] = strategy
                experiment_info = validate_autoai_experiment(experiment_info, self.pod_version)

                TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**experiment_info)

    def test_11_initialize_optimizer_kb_scenario_invalid_fields_set(self):
        with self.assertRaises(InvalidImputationParameterNonTS):
            experiment_info = self.experiment_info.copy()
            experiment_info['numerical_imputation_strategy'] = ImputationStrategy.MEAN
            experiment_info['numerical_imputation_value'] = 0
            experiment_info = validate_autoai_experiment(experiment_info, self.pod_version)

            TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**experiment_info)

        with self.assertRaises(InvalidImputationParameterNonTS):
            experiment_info = self.experiment_info.copy()
            experiment_info['numerical_imputation_strategy'] = ImputationStrategy.MEAN
            experiment_info['imputation_threshold'] = 0.1
            experiment_info = validate_autoai_experiment(experiment_info, self.pod_version)

            TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**experiment_info)

    def test_12_initialize_optimizer_ts_scenario(self):
        ts_experiment_info = self.ts_experiment_info.copy()

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0]\
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertIn('use_imputation', params)
        self.assertEqual(params['use_imputation'], True)
        self.assertIn('imputer_list', params)
        self.assertEqual(params['imputer_list'], ['FlattenIterative', 'Linear', 'Cubic', 'Previous'])
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_13_initialize_optimizer_ts_scenario_mean(self):
        ts_experiment_info = self.ts_experiment_info.copy()
        ts_experiment_info['numerical_imputation_strategy'] = ImputationStrategy.MEAN

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertIn('use_imputation', params)
        self.assertEqual(params['use_imputation'], True)
        self.assertIn('imputer_list', params)
        self.assertEqual(params['imputer_list'], ['Fill'])
        self.assertIn('imputer_fill_type', params)
        self.assertEqual(params['imputer_fill_type'], 'mean')
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_14_initialize_optimizer_ts_scenario_median(self):
        ts_experiment_info = self.ts_experiment_info.copy()
        ts_experiment_info['numerical_imputation_strategy'] = ImputationStrategy.MEDIAN

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertIn('use_imputation', params)
        self.assertEqual(params['use_imputation'], True)
        self.assertIn('imputer_list', params)
        self.assertEqual(params['imputer_list'], ['Fill'])
        self.assertIn('imputer_fill_type', params)
        self.assertEqual(params['imputer_fill_type'], 'median')
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_15_initialize_optimizer_ts_scenario_best_of_default_imputers(self):
        ts_experiment_info = self.ts_experiment_info.copy()
        ts_experiment_info['numerical_imputation_strategy'] = ImputationStrategy.BEST_OF_DEFAULT_IMPUTERS

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertIn('use_imputation', params)
        self.assertEqual(params['use_imputation'], True)
        self.assertIn('imputer_list', params)
        self.assertEqual(params['imputer_list'], ['FlattenIterative', 'Linear', 'Cubic', 'Previous'])
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_16_initialize_optimizer_ts_scenario_value(self):
        ts_experiment_info = self.ts_experiment_info.copy()
        ts_experiment_info['numerical_imputation_strategy'] = ImputationStrategy.VALUE

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertIn('use_imputation', params)
        self.assertEqual(params['use_imputation'], True)
        self.assertIn('imputer_list', params)
        self.assertEqual(params['imputer_list'], ['Fill'])
        self.assertIn('imputer_fill_type', params)
        self.assertEqual(params['imputer_fill_type'], 'value')
        self.assertIn('imputer_fill_value', params)
        self.assertEqual(params['imputer_fill_value'], 0)
        self.assertNotIn('imputation_threshold', params)

    def test_17_initialize_optimizer_ts_scenario_flatten_iterative(self):
        ts_experiment_info = self.ts_experiment_info.copy()
        ts_experiment_info['numerical_imputation_strategy'] = ImputationStrategy.FLATTEN_ITERATIVE

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertIn('use_imputation', params)
        self.assertEqual(params['use_imputation'], True)
        self.assertIn('imputer_list', params)
        self.assertEqual(params['imputer_list'], ['FlattenIterative'])
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_18_initialize_optimizer_ts_scenario_linear(self):
        ts_experiment_info = self.ts_experiment_info.copy()
        ts_experiment_info['numerical_imputation_strategy'] = ImputationStrategy.LINEAR

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertIn('use_imputation', params)
        self.assertEqual(params['use_imputation'], True)
        self.assertIn('imputer_list', params)
        self.assertEqual(params['imputer_list'], ['Linear'])
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_19_initialize_optimizer_ts_scenario_cubic(self):
        ts_experiment_info = self.ts_experiment_info.copy()
        ts_experiment_info['numerical_imputation_strategy'] = ImputationStrategy.CUBIC

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertIn('use_imputation', params)
        self.assertEqual(params['use_imputation'], True)
        self.assertIn('imputer_list', params)
        self.assertEqual(params['imputer_list'], ['Cubic'])
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_20_initialize_optimizer_ts_scenario_previous(self):
        ts_experiment_info = self.ts_experiment_info.copy()
        ts_experiment_info['numerical_imputation_strategy'] = ImputationStrategy.PREVIOUS

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertIn('use_imputation', params)
        self.assertEqual(params['use_imputation'], True)
        self.assertIn('imputer_list', params)
        self.assertEqual(params['imputer_list'], ['Previous'])
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_21_initialize_optimizer_ts_scenario_next(self):
        ts_experiment_info = self.ts_experiment_info.copy()
        ts_experiment_info['numerical_imputation_strategy'] = ImputationStrategy.NEXT

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertIn('use_imputation', params)
        self.assertEqual(params['use_imputation'], True)
        self.assertIn('imputer_list', params)
        self.assertEqual(params['imputer_list'], ['Next'])
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_22_initialize_optimizer_ts_scenario_no_imputation(self):
        ts_experiment_info = self.ts_experiment_info.copy()
        ts_experiment_info['numerical_imputation_strategy'] = ImputationStrategy.NO_IMPUTATION

        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

        params = self.remote_auto_pipelines._engine._wml_stored_pipeline_details['entity']['document']['pipelines'][0] \
            ['nodes'][0]['parameters']['optimization']

        print(json.dumps(params, indent=4))

        self.assertNotIn('preprocessor_cat_imp_strategy', params)
        self.assertNotIn('preprocessor_num_imp_strategy', params)
        self.assertIn('use_imputation', params)
        self.assertEqual(params['use_imputation'], False)
        self.assertNotIn('imputer_list', params)
        self.assertNotIn('imputer_fill_type', params)
        self.assertNotIn('imputer_fill_value', params)
        self.assertNotIn('imputation_threshold', params)

    def test_23_initialize_optimizer_ts_scenario_invalid_strategies(self):
        for strategy in [ImputationStrategy.MOST_FREQUENT]:
            with self.assertRaises(StrategyIsNotApplicable):
                ts_experiment_info = self.ts_experiment_info.copy()
                ts_experiment_info['numerical_imputation_strategy'] = strategy

                TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

    def test_24_initialize_optimizer_ts_scenario_invalid_fields_set(self):
        with self.assertRaises(NumericalImputationStrategyValueMisused):
            ts_experiment_info = self.ts_experiment_info.copy()
            ts_experiment_info['numerical_imputation_strategy'] = ImputationStrategy.MEAN
            ts_experiment_info['numerical_imputation_value'] = 0

            TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)

        with self.assertRaises(InvalidImputationParameterTS):
            ts_experiment_info = self.ts_experiment_info.copy()
            ts_experiment_info['categorical_imputation_strategy'] = ImputationStrategy.MEAN

            TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(**ts_experiment_info)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

import logging
from preparation_and_cleaning import *
from models_preparation import *


class TestAIFunction(unittest.TestCase):
    runtime_uid = None
    deployment_uid = None
    function_uid = None
    scoring_url = None
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestAIFunction.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.function_name = 'sample_function'
        self.deployment_name = "Test deployment"
        self.project_id = os.environ['PROJECT_ID']
        self.client.set.default_project(self.project_id)

    # def test_01_service_instance_details(self):
    #     TestAIFunction.logger.info("Check client ...")
    #     self.assertTrue(self.client.__class__.__name__ == 'APIClient')
    #
    #     TestAIFunction.logger.info("Getting instance details ...")
    #     details = self.client.service_instance.get_details()
    #     TestAIFunction.logger.debug(details)
    #
    #     self.assertTrue("published_models" in str(details))
    #     self.assertEqual(type(details), dict)

    def test_01_create_ai_function(self):

        self.client.repository.FunctionMetaNames.show()

        function_props = {
            self.client.repository.FunctionMetaNames.NAME: 'sample_function-sample_2',
            self.client.repository.FunctionMetaNames.RUNTIME_UID: 'ai-function_0.1-py3.6',
            self.client.repository.FunctionMetaNames.INPUT_DATA_SCHEMAS: [{
                "id": "iris function",
                "fields": [
                    {"metadata": {}, "type": "double", "name": "sepal length (cm)", "nullable": False},
                    {"metadata": {}, "type": "double", "name": "sepal width (cm)", "nullable": False},
                    {"metadata": {}, "type": "double", "name": "petal length (cm)", "nullable": False},
                    {"metadata": {}, "type": "double", "name": "petal width (cm)", "nullable": False}
                ]
            }],
            self.client.repository.FunctionMetaNames.OUTPUT_DATA_SCHEMAS: [{
                "id": "iris function",
                "fields": [
                    {"metadata": {'modeling_role': 'prediction'}, "type": "string", "name": "iris_type",
                     "nullable": False}
                ]
            }]

        }

        def score(payload):
            payload = payload['input_data'][0]
            values = [[row[0] * row[1]] for row in payload['values']]
            return {'predictions': [{'fields': ['multiplication'], 'values': values}]}

        ai_function_details = self.client.repository.store_function(score, function_props)

        TestAIFunction.function_uid = self.client.repository.get_function_uid(ai_function_details)
       # function_url = self.client.repository.get_function_href(ai_function_details)
        TestAIFunction.logger.info("AI function ID:" + str(TestAIFunction.function_uid))
       # TestAIFunction.logger.info("AI function URL:" + str(function_url))
        self.assertIsNotNone(TestAIFunction.function_uid)


    def test_03_download_ai_function_content(self):
        try:
            os.remove('test_ai_function.tar.gz')
        except:
            pass
        self.client.repository.download(TestAIFunction.function_uid, filename='test_ai_function_v42.gz')
        try:
            os.remove('test_ai_function.tar.gz')
        except:
            pass

    def test_04_get_details(self):

        details = self.client.repository.get_function_details(self.function_uid)
        print(details)
        self.assertTrue(self.function_name in str(details))


    def test_11_delete_function(self):
        TestAIFunction.logger.info("Delete function")
        self.client.repository.delete(TestAIFunction.function_uid)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import time
import joblib
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError

from ibmfl.util import config
from ibmfl.model.fl_model import FLModel
from ibmfl.exceptions import ModelInitializationException

logger = logging.getLogger(__name__)


class SklearnFLModel(FLModel):
    """
    Wrapper class for sklearn models.
    """

    def __init__(self, model_name, model_spec, sklearn_model=None, **kwargs):
        """
        Create a `SklearnFLModel` instance from a sklearn model.
        If sklearn_model is provided, it will use it; otherwise it will take
        the model_spec to create the model.

        :param model_name: A name specifying the type of model, e.g., \
                linear_SVM
        :type model_name: `str`
        :param model_spec: A dictionary contains model specification
        :type model_spec: `dict`
        :param sklearn_model: A compiled sklearn model
        :type sklearn_model: `sklearn`
        :param kwargs: A dictionary contains other parameter settings on \
         to initialize a sklearn model.
        :type kwargs: `dict`
        """
        super().__init__(model_name, model_spec, **kwargs)

        if sklearn_model is None:
            if model_spec is None or (type(model_spec) is not dict):
                raise ValueError('Initializing model requires '
                                 'a model specification or '
                                 'compiled sklearn model. '
                                 'None was provided.')
            self.model = self.load_model_from_spec(model_spec)
        else:
            self.model = sklearn_model

        self.is_classification = True if not(model_spec and model_spec.get(
            'is_classification')) else model_spec.get('is_classification')

    def fit_model(self, train_data, fit_params=None, **kwargs):
        """
        Fits current model with provided training data.

        :param train_data: Training data a tuple
        given in the form (x_train, y_train).
        :type train_data: `np.ndarray`
        :param fit_params: (Optional) Dictionary with hyperparameters that \
        will be used to train the model. \
        If the corresponding sklearn fit function is called, \
        the provided hyperparameter should only contains parameters that \
        match sklearn expected values, e.g., `learning_rate`, which provides \
        the learning rate schedule
        :return: None
        """
        raise NotImplementedError

    def update_model(self, model_update):
        """
        Update sklearn model with provided model_update.

        :param model_update: `ModelUpdate` object that contains \
        the required information to update the sklearn model.
        :type model_update: `ModelUpdate`
        :return: None
        """
        raise NotImplementedError

    def get_model_update(self):
        """
        Generates a `ModelUpdate` object that will be sent to other entities.

        :return: ModelUpdate
        :rtype: `ModelUpdate`
        """
        raise NotImplementedError

    def predict(self, x):
        """
        Perform prediction for the given input.

        :param x: Samples with shape as expected by the model.
        :type x: `np.ndarray`
        :return: Array of predictions
        :rtype: `np.ndarray`
        """
        return self.model.predict(x)

    def evaluate_model(self, x, y, **kwargs):
        """
        Evaluates the model given test data x and the corresponding labels y.

        :param x: Samples with shape as expected by the model.
        :type x: `np.ndarray`
        :param y: Corresponding true labels to x
        :type y: `np.ndarray`
        :param kwargs: Dictionary of model-specific arguments \
        for evaluating models. For example, sample weights accepted \
        by model.score.
        :return: Dictionary with all evaluation metrics provided by \
        specific implementation.
        :rtype: `dict`
        """
        raise NotImplementedError

    def save_model(self, filename=None, path=None):
        """
        Save a sklearn model to file in the format specific
        to the framework requirement.

        :param filename: Name of the file where to store the model.
        :type filename: `str`
        :param path: Path of the folder where to store the model. If no path \
        is specified, the model will be stored in the default data location of \
        the library `DATA_PATH`.
        :type path: `str`
        :return: filename
        """
        if filename is None:
            file = self.model_name if self.model_name else self.model_type
            filename = '{}_{}.pickle'.format(file, time.time())

        full_path = super().get_model_absolute_path(filename)

        with open(full_path, 'wb') as f:
            joblib.dump(self.model, f)
        logger.info('Model saved in path: %s.', full_path)
        return filename

    @staticmethod
    def load_model(file_name):
        """
        Load a sklearn from the disk given the filename.

        :param file_name: Name of the file that contains the model to be loaded.
        :type file_name: `str`
        :return: A sklearn model
        :rtype: `sklearn`
        """
        absolute_path = config.get_absolute_path(file_name)

        with open(absolute_path, 'rb') as f:
            model = joblib.load(f)

        return model

    @staticmethod
    def load_model_from_spec(model_spec):
        """
        Loads model from provided model_spec, where model_spec is a `dict`
        that contains the following items: model_spec['model_definition']
        contains the model definition as a specific sklearn model type,
        e.g., `sklearn.linear_model.SGDClassifier`.

        :param model_spec: Model specification contains \
        a compiled sklearn model.
        :type model_spec: `dict`
        :return: model
        :rtype: `sklearn`
        """
        model = None
        try:
            if 'model_definition' in model_spec:
                path = model_spec['model_definition']
                with open(path, 'rb') as f:
                    model = joblib.load(f)
        except Exception as ex:
            raise ModelInitializationException('Model specification was '
                                               'badly formed. '+ str(ex))
        return model

    def is_fitted(self):
        """
        Return a boolean value indicating if the model is fitted or not.

        :return: res
        :rtype: `bool`
        """
        try:
            check_is_fitted(self.model)
        except NotFittedError:
            return False
        return True
